#!/usr/bin/env python3
import tkinter as tk
import sys
import socket
import socketserver as SocketServer

import os
import select
import getopt
import time
import glob

import paramiko
import threading
import queue

import yaml

from urllib.parse import urlparse, parse_qs
from os.path import expanduser
from datetime import datetime

from .key import Key, read_key

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

HOST = "127.0.0.1"
PORT = 65432
SSH_PORT = 22
REMOTE_CONFIG_FILE = "~/.config/hpc-campaign/hosts.yaml"

SSH_CONNECT_ERROR = 101
SSH_TUNNEL_ERROR = 102
SSH_NOPORT_ERROR = 103
SSH_NO_ERROR = 0

SUPPORTED_PROTOCOLS = ["ssh", "ssh_tunnel"]
SUPPORTED_AUTH_METHODS = ["passcode", "password", "publickey"]
THREAD_CHECK_TIME = 0.5  # 0.5 seconds


class LoginWindowRemote(object):
    def __init__(self, master, remote, username, auth):
        global g_remote_user, g_remote_pass
        g_remote_user = None
        g_remote_pass = None
        self.master = master

        # Remote host
        self.rl = tk.Label(master, text="Remote Host: " + remote)
        self.rl.config(font=("Courier", 14))
        self.rl.pack()
        self.rl1 = tk.Label(master, text="Username")
        self.rl1.config(font=("Courier", 14))
        self.rl1.pack()
        self.rb = tk.Entry(master, font=("Century 12"), width=64)
        if username is not None:
            self.rb.insert(0, username)
        self.rb.pack()
        self.rl1 = tk.Label(master, text=auth)
        self.rl1.config(font=("Courier", 14))
        self.rl1.pack()
        self.rb1 = tk.Entry(master, show="*", font=("Century 12"), width=64)
        self.rb1.pack()

        master.bind("<Return>", self.get_value)
        self.b2 = tk.Button(master, text="Enter", command=self.get_value)
        self.b2.pack()

    def get_value(self, event=None):
        global g_remote_user, g_remote_pass
        g_remote_user = self.rb.get()
        g_remote_pass = self.rb1.get()
        # print("g_user:",g_user,g_pass)
        self.destroy()

    def destroy(self):
        print("Destroying window.")
        self.master.withdraw()
        self.master.quit()
        self.master.destroy()
        print("Done!")


class LoginWindowJumpRemote(object):
    def __init__(self, master, jump_host, jump_user, jump_auth, remote_host, remote_user, remote_auth):
        global g_remote_user, g_remote_pass
        global g_jump_user, g_jump_pass
        g_remote_user = None
        g_remote_pass = None
        g_jump_user = None
        g_jump_pass = None
        self.master = master

        # Jump host
        self.jl = tk.Label(master, text="Jump Host: " + jump_host)
        self.jl.config(font=("Courier", 14))
        self.jl.pack()
        self.jl1 = tk.Label(master, text="Jump Username")
        self.jl1.config(font=("Courier", 14))
        self.jl1.pack()
        self.jb = tk.Entry(master, font=("Century 12"), width=64)
        if jump_user is not None:
            self.jb.insert(0, jump_user)
        self.jb.pack()
        self.jl1 = tk.Label(master, text=jump_auth)
        self.jl1.config(font=("Courier", 14))
        self.jl1.pack()
        self.jb1 = tk.Entry(master, show="*", font=("Century 12"), width=64)
        self.jb1.pack()

        # Remote host
        self.rl = tk.Label(master, text="Remote Host: " + remote_host)
        self.rl.config(font=("Courier", 14))
        self.rl.pack()
        self.rl1 = tk.Label(master, text="Remote Username")
        self.rl1.config(font=("Courier", 14))
        self.rl1.pack()
        self.rb = tk.Entry(master, font=("Century 12"), width=64)
        if remote_user is not None:
            self.rb.insert(0, remote_user)
        self.rb.pack()
        self.rl1 = tk.Label(master, text=remote_auth)
        self.rl1.config(font=("Courier", 14))
        self.rl1.pack()
        self.rb1 = tk.Entry(master, show="*", font=("Century 12"), width=64)
        self.rb1.pack()

        master.bind("<Return>", self.get_value)
        self.b2 = tk.Button(master, text="Enter", command=self.get_value)
        self.b2.pack()

    def get_value(self, event=None):
        global g_remote_user, g_remote_pass
        global g_jump_user, g_jump_pass
        g_jump_user = self.jb.get()
        g_jump_pass = self.jb1.get()
        g_remote_user = self.rb.get()
        g_remote_pass = self.rb1.get()
        # print("g_user:",g_user,g_pass)
        self.destroy()

    def destroy(self):
        print("Destroying window.")
        self.master.withdraw()
        self.master.quit()
        self.master.destroy()
        print("Done!")


g_queue = queue.Queue()

# BEGIN
# From forward.py in paramiko
# https://github.com/paramiko/paramiko/blob/main/demos/forward.py
#
g_verbose = True


def verbose(s):
    if g_verbose:
        print(s)


class ForwardServer(SocketServer.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(SocketServer.BaseRequestHandler):
    def handle(self):
        global g_queue
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            verbose("Incoming request to %s:%d failed: %s" % (self.chain_host, self.chain_port, repr(e)))
            return
        if chan is None:
            verbose("Incoming request to %s:%d was rejected by the SSH server." % (self.chain_host, self.chain_port))
            return

        verbose(
            "Connected!  Tunnel open %r -> %r -> %r"
            % (
                self.request.getpeername(),
                chan.getpeername(),
                (self.chain_host, self.chain_port),
            )
        )
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        verbose("Tunnel closed from %r" % (peername,))
        print("SELF TEMP_TUNNEL: ", self.temp_tunnel)
        if self.temp_tunnel is True:
            g_queue.put((self.chain_host, self.chain_port))


#
# From forward.py in paramiko
# END


#
# SSH tunnel implementation based on forward.py in paramiko
#
class SSHOptions:
    def __init__(self, keyfile=None, look_for_keys=True):
        self.keyfile = keyfile
        self.look_for_keys = look_for_keys

    def set_options(self, keyfile=None, look_for_keys=True):
        self.keyfile = keyfile
        self.look_for_keys = look_for_keys


class SSHServerInfo:
    def __init__(self, host_name, host_port):
        self.host_name = host_name
        self.host_port = host_port

    def set_server(self, host_name, host_port):
        self.host_name = host_name
        self.host_port = host_port


class SSHUserInfo:
    def __init__(self, user_name, user_pass):
        self.user_name = user_name
        self.user_pass = user_pass

    def set_user(self, user_name, user_pass):
        self.user_name = user_name
        self.user_pass = user_pass


class SSHJumpLinkInfo:
    def __init__(self, transport, channel):
        self.transport = transport
        self.channel = channel

    def set_jump_link(self, transport, channel):
        self.transport = transport
        self.channel = channel


class SSHConnectedServerInfo:
    def __init__(self, server: SSHServerInfo, user: SSHUserInfo, client: paramiko.SSHClient):
        self.server = server
        self.user = user
        self.client = client

    def set_server_info(self, server: SSHServerInfo, user: SSHUserInfo, client: paramiko.SSHClient):
        self.server = server
        self.user = user
        self.client = client


class SSHRemoteConnectionInfo:
    def __init__(
        self,
        remote: SSHConnectedServerInfo,
        jump: SSHConnectedServerInfo,
        jump_link: SSHJumpLinkInfo,
    ):
        self.remote = remote
        self.jump = jump
        self.jump_link = jump_link

    def set_remote_connection_info(
        self,
        remote: SSHConnectedServerInfo,
        jump: SSHConnectedServerInfo,
        jump_link: SSHJumpLinkInfo,
    ):
        self.remote = remote
        self.jump = jump
        self.jump_link = jump_link


class SSHTunnelInfo:
    def __init__(
        self,
        ssh_client=None,
        service_group=None,
        service_name=None,
        jump_host=None,
        remote_host=None,
        remote_port=None,
        local_port=None,
        cookie=None,
        thread_process=None,
        forward_server=None,
        forward_tunnel=False,
        dest_host=None,
        dest_port=None,
        reverse_tunnel=False,
    ):
        self.ssh_client = ssh_client
        self.service_group = service_group
        self.service_name = service_name
        self.jump_host = jump_host
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.dest_host = dest_host
        self.dest_port = dest_port
        self.local_port = local_port
        self.cookie = cookie
        self.process = thread_process
        self.forward_server = forward_server
        self.forward_tunnel = forward_tunnel

        self.reverse_tunnel = reverse_tunnel

        if self.forward_server is not None:
            self.forward_tunnel = True

    def set_tunnel_info(
        self,
        ssh_client=None,
        service_group=None,
        service_name=None,
        jump_host=None,
        remote_host=None,
        remote_port=None,
        local_port=None,
        cookie=None,
        thread_process=None,
        forward_server=None,
        forward_tunnel=False,
        dest_host=None,
        dest_port=None,
        reverse_tunnel=False,
    ):
        self.ssh_client = ssh_client
        self.service_group = service_group
        self.service_name = service_name
        self.jump_host = jump_host
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.dest_host = dest_host
        self.dest_port = dest_port
        self.local_port = local_port
        self.cookie = cookie
        self.process = thread_process
        self.forward_server = forward_server
        self.forward_tunnel = forward_tunnel
        self.reverse_tunnel = reverse_tunnel

        if self.forward_server is not None:
            self.forward_tunnel = True


g_proxy_command = None


class SSHConnectRemote:
    def __init__(self):
        self.options = SSHOptions()
        self.jump = None
        self.jump_transport = None
        self.jump_channel = None
        self.remote = None
        self.remote_connection_info = None

    def _connect_server(
        self,
        host_name: str,
        ssh_port: int = SSH_PORT,
        user_name: str = None,
        user_pass: str = None,
        sock_channel: paramiko.Channel = None,
    ):
        verbose("Connecting to remote server %s:%d ..." % (host_name, ssh_port))

        # Connect to jump server
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if g_proxy_command:
            proxy_jump_command = g_proxy_command.replace("%h", host_name).replace("%p", str(ssh_port))
            # verbose("    using proxy command:", proxy_jump_command)
            proxy = paramiko.ProxyCommand(proxy_jump_command)
            sock_channel = proxy

        try:
            client.connect(
                hostname=host_name,
                port=ssh_port,
                username=user_name,
                key_filename=self.options.keyfile,
                look_for_keys=self.options.look_for_keys,
                password=user_pass,
                sock=sock_channel,
            )
        except Exception as e:
            print("*** Failed to connect to %s:%d: %r" % (host_name, ssh_port, e))
            return SSH_CONNECT_ERROR, None

        server = SSHServerInfo(host_name, ssh_port)
        user = SSHUserInfo(user_name, user_pass)
        remote = SSHConnectedServerInfo(server, user, client)

        return SSH_NO_ERROR, remote

    def connect_remote(
        self,
        host_name: str = None,
        ssh_port: int = SSH_PORT,
        user_name: str = None,
        user_pass: str = None,
        sock_channel: paramiko.Channel = None,
    ):

        conn_err, self.remote = self._connect_server(
            host_name, ssh_port, user_name, user_pass, sock_channel=sock_channel
        )
        if conn_err != SSH_NO_ERROR:
            return SSH_CONNECT_ERROR

        self.remote_connection_info = SSHRemoteConnectionInfo(self.remote, jump=None, jump_link=None)
        return SSH_NO_ERROR

    def connect_remote_via_jump(
        self,
        connected_jump: SSHConnectedServerInfo,
        remote_host_name: str,
        remote_ssh_port: int,
        remote_user_name: str,
        remote_user_pass: int,
    ):

        verbose(
            "Connecting to remote server %s via connected jump server %s."
            % (remote_host_name, connected_jump.server.host_name)
        )

        try:
            # Create a channel via jump server
            self.jump_transport = connected_jump.client.get_transport()
            dest_addr = (remote_host_name, remote_ssh_port)
            src_addr = (connected_jump.server.host_name, connected_jump.server.host_port)
            self.jump_channel = self.jump_transport.open_channel("direct-tcpip", dest_addr, src_addr)
        except Exception as e:
            print("An exception has occured: %r" % (e))
            return None

        # Connect to remote server (destination for port forwarding)
        conn_err, self.remote = self._connect_server(
            host_name=remote_host_name,
            ssh_port=remote_ssh_port,
            user_name=remote_user_name,
            user_pass=remote_user_pass,
            sock_channel=self.jump_channel,
        )
        if conn_err != SSH_NO_ERROR:
            return SSH_CONNECT_ERROR

        self.jump = connected_jump
        jump_link = SSHJumpLinkInfo(self.jump_transport, self.jump_channel)
        self.remote_connection_info = SSHRemoteConnectionInfo(self.remote, self.jump, jump_link)

        return SSH_NO_ERROR

    def get_remote_connection_info(self):
        return self.remote_connection_info


class SSHLocalRemoteTunnel:
    def __init__(self, remote_connection: SSHConnectRemote = None):
        # Connected remote server for local or reverse port forwarding
        self.remote_connection = remote_connection
        self.remote_connection_info = remote_connection.get_remote_connection_info()
        # Port forwarding servers
        self.forward_server = None
        self.local_port = None

    def set_remote_connection(self, remote_connection: SSHConnectRemote):
        self.remote_connection = remote_connection
        self.remote_connection_info = remote_connection.get_remote_connection_info()

    def get_remote_connection(self):
        return self.remote_connection

    def get_remote_connection_info(self):
        return self.remote_connection_info

    def exec_remote_command(self, service_command: str):
        client = self.remote_connection_info.remote.client
        return client.exec_command(service_command, get_pty=False)

    def _forward_tunnel(
        self,
        local_port: int,
        dest_host: str,
        dest_port: int,
        transport: paramiko.Transport,
        temptunnel: bool,
    ):
        class SubHandler(Handler):
            chain_host = dest_host
            chain_port = dest_port
            ssh_transport = transport
            temp_tunnel = temptunnel

        forward_server = ForwardServer(("", local_port), SubHandler)
        return forward_server

    def _ssh_forward(self, client: paramiko.SSHClient, local_port: int, dest: SSHServerInfo, temptunnel: bool):
        verbose("Opening tunnel for local port %d to %s:%d " % (local_port, dest.host_name, dest.host_port))
        forward_server = None
        try:
            forward_server = self._forward_tunnel(
                local_port, dest.host_name, dest.host_port, client.get_transport(), temptunnel
            )
            return forward_server
        except Exception as e:
            print("An exception has occured: %r" % (e))
            return None

    def _run_forward_server(self, forward_server):
        verbose("Starting.")
        forward_server.serve_forever()
        verbose("Shut down.")

    def start_forward_server(self, local_port: int, dest_server: SSHServerInfo, temptunnel: bool):
        client = self.remote_connection_info.remote.client
        self.forward_server = self._ssh_forward(client, local_port, dest_server, temptunnel)

        if self.forward_server is None:
            return None, None, None, SSH_TUNNEL_ERROR
        verbose("Got the forward server")

        thread = threading.Thread(target=self._run_forward_server, args=(self.forward_server,))
        thread.daemon = True
        thread.start()
        while thread.is_alive() is False:
            time.sleep(THREAD_CHECK_TIME)

        return self.forward_server, thread, SSH_NO_ERROR

    def _reverse_forward_handler(self, chan: paramiko.Channel, host: str, port: int):
        sock = socket.socket()
        try:
            sock.connect((host, port))
        except Exception as e:
            verbose("Forwarding request to %s:%d failed: %r" % (host, port, e))
            return

        verbose("Connected!  Tunnel open %r -> %r -> %r" % (chan.origin_addr, chan.getpeername(), (host, port)))

        while True:
            r, w, x = select.select([sock, chan], [], [])
            if sock in r:
                data = sock.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                sock.send(data)
        chan.close()
        sock.close()
        verbose("Tunnel closed from %r" % (chan.origin_addr,))

    def _reverse_forward_tunnel(self, remote_server_port: int, dest_host: str, dest_port: int):
        client = self.remote_connection_info.remote.client
        transport = client.get_transport()
        transport.request_port_forward("", remote_server_port)
        while True:
            chan = transport.accept(1000)
            if chan is None:
                continue
            thread = threading.Thread(target=self._reverse_forward_handler, args=(chan, dest_host, dest_port))
            thread.daemon = True
            thread.start()
            while thread.is_alive() is False:
                time.sleep(THREAD_CHECK_TIME)

    def reverse_forward_tunnel(self, remote_server_port: int, dest_host: str, dest_port: int):
        thread = threading.Thread(target=self._reverse_forward_tunnel, args=(remote_server_port, dest_host, dest_port))
        thread.deamon = True
        thread.start()
        while thread.is_alive() is False:
            time.sleep(THREAD_CHECK_TIME)
        return thread

    def shutdown_forward_server(self):
        verbose("Shutting down tunnel.")
        self.forward_server.shutdown()


g_server_config_data = None


def read_yaml_config(filename: str):
    config_data = None
    if filename is not None:
        f = open(filename, mode="r")
        config_data = yaml.safe_load(f)
        f.close()
    return config_data


g_open_tunnel_list: list[SSHTunnelInfo] = []
g_remote_conn_list: list[SSHConnectRemote] = []


class MyTCPHandler(SocketServer.BaseRequestHandler):
    def handle(self):
        # self.server_config_data = g_server_config_data
        req_data = "".join(self.request.recv(2048).decode("UTF-8").split())
        req_data = "".join(req_data.rsplit("\x00"))
        print("Client {}:".format(self.client_address[0]))
        print("Request  :", req_data)

        req_path, req_qry = self.parse_request(req_data)
        if not req_qry:
            self.send_response("port:-1,msg:incorrect_request_format")
            return
        print("Parsed Request: ", req_qry)

        if req_path == "/run_service":
            self.do_run_remote_service(req_qry)
            return
        elif req_path == "/connect_port":
            # /connect_port?jhost=<jumphost>&juser=<jumpuser>&rhost=<remotehost>&ruser=<remoteuser>&dhost=<desthost>&dport=<destport>
            # /connect_port?group=<group_name>&service=<service_name>&dhost=<desthost>&dport=<destport>
            self.do_connect_port(req_qry)
        elif req_path == "/reverse":
            self.do_reverse_port_forward(req_qry)
            return
        elif req_path == "/get_key":
            self.send_key(req_qry)
            return
        else:
            self.send_response("port:-1,msg:incorrect_request_format")
            return

    def parse_request(self, req_data):
        req_tmp = urlparse(req_data)
        req_qry = parse_qs(req_tmp.query)
        return req_tmp.path, req_qry

    def send_response(self, res_str):
        self.request.sendall(res_str.encode("UTF-8"))
        return

    def login_window_remote(
        self,
        try_cnt: int,
        remote_host_name: str,
        remote_user_name: str = None,
        auth_type: str = "Password",
    ):
        root = tk.Tk()
        tmp_str = "Login Window Remote"
        if try_cnt != 0:
            tmp_str = "Failed (" + str(try_cnt) + "/3). Try Again."
        root.title(tmp_str)
        _ = LoginWindowRemote(root, remote_host_name, remote_user_name, auth_type)
        root.mainloop()
        remote_user = SSHUserInfo(g_remote_user, g_remote_pass)
        return remote_user

    def login_window_jump_remote(
        self,
        try_cnt: int,
        jump_host_name: str,
        remote_host_name: str,
        jump_user_name: str = None,
        remote_user_name: str = None,
        jump_auth: str = "Password",
        remote_auth: str = "Password",
    ):

        root = tk.Tk()
        tmp_str = "Login Window Jump Remote"
        if try_cnt != 0:
            tmp_str = "Failed (" + str(try_cnt) + "/3). Try Again."
        root.title(tmp_str)
        _ = LoginWindowJumpRemote(
            root,
            jump_host_name,
            jump_user_name,
            jump_auth,
            remote_host_name,
            remote_user_name,
            remote_auth,
        )
        root.mainloop()
        jump_user = SSHUserInfo(g_jump_user, g_jump_pass)
        remote_user = SSHUserInfo(g_remote_user, g_remote_pass)
        return jump_user, remote_user

    def check_connected_remote(self, remote_host_name: str, remote_user_name: str) -> SSHConnectRemote:
        for srvr in g_remote_conn_list:
            srvr_info = srvr.get_remote_connection_info()
            verbose(
                "Checking if %s %s exists: %s %s"
                % (
                    remote_host_name,
                    remote_user_name,
                    srvr_info.remote.server.host_name,
                    srvr_info.remote.user.user_name,
                )
            )
            if (srvr_info.remote.server.host_name == remote_host_name) and (
                srvr_info.remote.user.user_name == remote_user_name
            ):
                return srvr
        return None

    def login_connect_remote(self, req_def) -> SSHConnectRemote:
        global g_remote_conn_list
        remote_host_name = req_def["remote_host"]
        remote_user_name = req_def["username"]
        ssh_connected_remote = self.check_connected_remote(remote_host_name, remote_user_name)
        if ssh_connected_remote is not None:
            return ssh_connected_remote

        jump_host_name = req_def["jumphost"]
        jump_user_name = req_def["jumpuser"]
        ssh_connected_jump = None
        if jump_host_name is not None:
            ssh_connected_jump = self.check_connected_remote(jump_host_name, jump_user_name)

        if req_def["auth"] == "publickey":
            ssh_connect_remote = SSHConnectRemote()
            if req_def["identity_file"] is not None:
                ssh_connect_remote.options.keyfile = expanduser(req_def["identity_file"])
            error_no = ssh_connect_remote.connect_remote(
                host_name=remote_host_name,
                ssh_port=SSH_PORT,
                user_name=remote_user_name,
                user_pass=None,
                sock_channel=None,
            )
            if error_no == SSH_NO_ERROR:
                g_remote_conn_list.append(ssh_connect_remote)
                return ssh_connect_remote
            else:
                return None

        try_connect_cnt = 0
        ssh_connect_remote = None
        ssh_connect_jump = None
        while try_connect_cnt <= 3:
            ssh_connect_remote = SSHConnectRemote()
            if ssh_connected_jump is not None:
                remote_user = self.login_window_remote(
                    try_connect_cnt,
                    remote_host_name,
                    remote_user_name,
                    req_def["auth"].capitalize(),
                )
                if remote_user.user_name is None:
                    ssh_connected = False
                    break
                jump_info = ssh_connected_jump.get_remote_connection_info()
                error_no = ssh_connect_remote.connect_remote_via_jump(
                    connected_jump=jump_info.remote,
                    remote_host_name=remote_host_name,
                    remote_ssh_port=SSH_PORT,
                    remote_user_name=remote_user.user_name,
                    remote_user_pass=remote_user.user_pass,
                )
            elif jump_host_name is not None:
                ssh_connect_jump = SSHConnectRemote()
                jump_user, remote_user = self.login_window_jump_remote(
                    try_connect_cnt,
                    jump_host_name,
                    remote_host_name,
                    jump_user_name,
                    remote_user_name,
                    req_def["auth"].capitalize(),
                    req_def["auth"].capitalize(),
                )
                if remote_user.user_name is None:
                    ssh_connected = False
                    break
                # Connect to the jump host
                error_no = ssh_connect_jump.connect_remote(
                    host_name=jump_host_name,
                    ssh_port=SSH_PORT,
                    user_name=jump_user.user_name,
                    user_pass=jump_user.user_pass,
                    sock_channel=None,
                )
                if error_no != SSH_NO_ERROR:
                    ssh_connected = False
                    break
                # Connect to remote via connected jump host
                error_no = ssh_connect_remote.connect_remote_via_jump(
                    connected_jump=ssh_connect_jump.remote,
                    remote_host_name=remote_host_name,
                    remote_ssh_port=SSH_PORT,
                    remote_user_name=remote_user.user_name,
                    remote_user_pass=remote_user.user_pass,
                )
            else:
                remote_user = self.login_window_remote(
                    try_connect_cnt,
                    remote_host_name,
                    remote_user_name,
                    req_def["auth"].capitalize(),
                )
                if remote_user.user_name is None:
                    ssh_connected = False
                    break
                error_no = ssh_connect_remote.connect_remote(
                    host_name=remote_host_name,
                    ssh_port=SSH_PORT,
                    user_name=remote_user.user_name,
                    user_pass=remote_user.user_pass,
                    sock_channel=None,
                )

            if error_no == SSH_NO_ERROR:
                ssh_connected = True
                break
            try_connect_cnt = try_connect_cnt + 1

        if ssh_connected is True:
            g_remote_conn_list.append(ssh_connect_remote)
            if ssh_connect_jump is not None:
                g_remote_conn_list.append(ssh_connect_jump)
            return ssh_connect_remote
        else:
            return None

    def check_requested_local_port_available(self, requested_port, max_count=200):
        verbose("Checking if port %d is available." % (requested_port))
        result = False
        try_cnt = 0
        while (result is False) and (try_cnt < max_count):
            a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            location = ("127.0.0.1", requested_port)
            try:
                a_socket.bind(location)
                result = True
            except socket.error:
                result = False
                try_cnt = try_cnt + 1
            a_socket.close()

        if result is False:
            verbose("Port %s is not available after %d tries" % (requested_port, self.max_count))
            return False
        else:
            return True

    def find_any_available_local_port(self, port_start=28000, max_count=2000):
        a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        location = ("", 0)
        assigned_port = None
        host_address = None
        try:
            a_socket.bind(location)
            (host_address, assigned_port) = a_socket.getsockname()
            a_socket.close()
        except socket.error as msg:
            verbose(f"Couldn't find any open ports because of Socket Error: {msg}")
            return None
        return assigned_port

    def find_available_local_port(self, port_start=28000, max_count=2000):
        result = False
        try_cnt = 0
        assigned_port = port_start
        while (result is False) and (try_cnt < max_count):
            verbose("Checking if port %d is available." % (assigned_port))
            a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            location = ("127.0.0.1", assigned_port)
            try:
                a_socket.bind(location)
                result = True
            except socket.error as msg:
                verbose(f"Socket Error: {msg}")
                result = False
                assigned_port = assigned_port + 1
                try_cnt = try_cnt + 1
            a_socket.close()

        if result is False:
            verbose("Couldn't find any open ports after %d tries" % (max_count))
            return None
        else:
            return assigned_port

    def get_local_port(self, requested_port=None, max_count=2000):
        self.local_port = requested_port
        if requested_port is not None:
            if self.check_requested_local_port_available(requested_port=requested_port, max_count=max_count) is False:
                self.local_port = None
        else:
            self.local_port = self.find_available_local_port()
        return self.local_port

    def check_service_running(self, service_name, service_group):
        for tnnl in g_open_tunnel_list:
            if (service_name == tnnl.service_name) and (service_group == tnnl.service_group):
                if tnnl.cookie is not None:
                    return tnnl.local_port, tnnl.cookie
                return tnnl.local_port, None
        return None, None

    def parse_service_request(self, req_qry):
        if "version" not in g_server_config_data:
            verbose("No config version, assuming v0.1")
            return self.parse_service_request_version01(req_qry)
        if g_server_config_data["version"] == "v0.1":
            verbose("Config version v0.1")
            return self.parse_service_request_version01(req_qry)
        verbose("Unsupported configuration version: %s" % (g_server_config_data["version"],))
        return None

    def parse_service_request_version01(self, req_qry):
        req_group_name = req_qry["group"][0]
        req_service_name = req_qry["service"][0]
        for group_name in g_server_config_data:
            if group_name == req_group_name:
                service_group = g_server_config_data[group_name]
                for service_name in service_group:
                    if service_name == req_service_name:
                        service_data = service_group[service_name]
                        if (
                            ("serverpath" not in service_data)
                            or ("host" not in service_data)
                            or ("protocol" not in service_data)
                            or ("authentication" not in service_data)
                        ):
                            return None
                        req_def = {}
                        req_def["service_group"] = req_group_name
                        req_def["service_name"] = service_name
                        req_def["service_prog"] = service_data["serverpath"]
                        req_def["service_args"] = None
                        if "args" in service_data:
                            req_def["service_args"] = service_data["args"]
                        req_def["username"] = None
                        if "user" in service_data:
                            req_def["username"] = service_data["user"]
                        req_def["remote_host"] = service_data["host"]
                        req_def["dest_host"] = service_data["host"]
                        req_def["jumphost"] = None
                        if "jumphost" in service_data:
                            req_def["jumphost"] = service_data["jumphost"]
                        req_def["jumpuser"] = None
                        if "jumpuser" in service_data:
                            req_def["jumpuser"] = service_data["jumpuser"]
                        if "port" in service_data:
                            req_def["remote_port"] = int(service_data["port"])
                        else:
                            req_def["remote_port"] = -1
                        req_def["tunneltype"] = "permanent"
                        if "tunneltype" in service_data:
                            req_def["tunneltype"] = service_data["tunneltype"]
                        req_def["conn"] = service_data["protocol"]
                        req_def["auth"] = service_data["authentication"]
                        if "identity_file" in service_data:
                            req_def["identity_file"] = service_data["identity_file"]
                        if ("local_port" not in req_qry) and ("local_port" in service_data):
                            req_def["local_port"] = service_data["local_port"]
                        elif ("local_port" in req_qry) and (int(req_qry["local_port"][0]) > 0):
                            req_def["local_port"] = req_qry["local_port"][0]
                        return req_def
        return None

    def do_run_remote_service(self, req_qry):
        print("Remote service request: ", req_qry)
        if "group" not in req_qry:
            self.send_response("port:-1,msg:missing_group_in_request")
            return
        if "service" not in req_qry:
            self.send_response("port:-1,msg:missing_service_in_request")
            return

        local_port, service_cookie = self.check_service_running(
            service_name=req_qry["service"][0], service_group=req_qry["group"][0]
        )
        if local_port is not None:
            if service_cookie is not None:
                self.send_response("port:" + str(local_port) + ",cookie:" + str(service_cookie) + ",msg:no_error")
            else:
                self.send_response("port:" + str(local_port) + ",msg:no_error")
            return

        req_def = self.parse_service_request(req_qry)
        if req_def is None:
            self.send_response("port:-1,msg:incomplete_service_definition")
            return

        print("Service request: ", req_def)
        self.run_service(req_def)
        return

    def parse_service_response(self, r_stdout, r_stderr):
        print("Parsing service response...")
        service_data = {}
        # for line in r_stdout.readlines():  -- replaced it with the following
        # two lines. Also commented out the break statement.
        line = r_stdout.readline()
        if line != "":
            print("LINE: ", line)
            srvr_line = line.replace(" ", "")
            srvr_line = srvr_line.replace("\r", "")
            srvr_line = srvr_line.replace("\n", "")
            if ("port:" in srvr_line) and ("msg:" in srvr_line):
                srv_str = srvr_line.replace(" ", "")
                srv_str = srv_str.replace("\n", "")
                srv_str = srv_str.split(";")
                for srv_item in srv_str:
                    tmp_str = srv_item.split(":", maxsplit=1)
                    service_data[tmp_str[0]] = tmp_str[1]
                # break
        if not service_data:
            # for line in r_stderr.readline(): -- replaced it with the
            # following two lines. Also commented out the break statement.
            line = r_stderr.readline()
            if line != "":  # for line in r_stderr.readline():
                print("LINE: ", line)
                srvr_line = line.replace(" ", "")
                srvr_line = srvr_line.replace("\r", "")
                srvr_line = srvr_line.replace("\n", "")
                if ("port:" in srvr_line) and ("msg:" in srvr_line):
                    srv_str = srvr_line.replace(" ", "")
                    srv_str = srv_str.replace("\n", "")
                    srv_str = srv_str.split(";")
                    for srv_item in srv_str:
                        tmp_str = srv_item.split(":", maxsplit=1)
                        service_data[tmp_str[0]] = tmp_str[1]
                    # break
        if not service_data:
            return None
        if "cookie" not in service_data:
            service_data["cookie"] = None
        return service_data

    def run_service(self, req_def):
        global g_open_tunnel_list
        print("Run Service Request:", req_def)

        if (
            ("service_prog" not in req_def)
            or ("remote_host" not in req_def)
            or ("conn" not in req_def)
            or ("auth" not in req_def)
        ):
            self.send_response("port:-1,msg:incorrect_service_request_format")
            return
        if not req_def["conn"] in SUPPORTED_PROTOCOLS:
            self.send_response("port:-1,msg:unsupported_protocol")
            return
        if not req_def["auth"] in SUPPORTED_AUTH_METHODS:
            self.send_response("port:-1,msg:unsupported_authication_method")
            return

        requested_port = None
        if "local_port" in req_def:
            requested_port = int(req_def["local_port"])
            if self.check_requested_local_port_available(requested_port) is False:
                self.send_response("port:" + str(requested_port) + ",msg:requested_port_is_not_available")
                return

        ssh_connect_remote = self.login_connect_remote(req_def)
        if ssh_connect_remote is None:
            self.send_response("port:-1,msg:cannot_connect_to_remote_server")
            return

        ssh_tunnel_client = SSHLocalRemoteTunnel(ssh_connect_remote)

        # Run the remote service
        service_command = req_def["service_prog"]
        if req_def["service_args"] is not None:
            service_command = service_command + " " + req_def["service_args"]
        print("Service command:", service_command)
        r_stdin, r_stdout, r_stderr = ssh_tunnel_client.exec_remote_command(service_command)
        service_data = self.parse_service_response(r_stdout, r_stderr)
        print("SERVICE DATA: ", service_data)
        if service_data is None:
            self.send_response("port:-1,msg:service_not_available")
            return
        print("Service data:", service_data)
        service_data["port"] = int(service_data["port"])
        if service_data["port"] < 0:
            self.send_response("port:-1,msg:cannot_find_service_port_on_remote")
            return

        service_name = req_def["service_name"]
        service_group = req_def["service_group"]
        remote_host = req_def["remote_host"]
        remote_port = service_data["port"]
        dest_host = req_def["dest_host"]
        dest_port = service_data["port"]
        remote_srvr = SSHServerInfo(dest_host, dest_port)

        local_port = self.get_local_port(requested_port)
        if local_port is None:
            self.send_response("port:-1,msg:cannot_find_local_port")
            return

        temptunnel = False
        if "transient" in req_def["tunneltype"]:
            temptunnel = True

        forward_server, thread_process, error_no = ssh_tunnel_client.start_forward_server(
            local_port, remote_srvr, temptunnel=temptunnel
        )
        if error_no == SSH_NO_ERROR:
            remote_connect_info = ssh_tunnel_client.get_remote_connection_info()
            jump_host = None
            if remote_connect_info.jump is not None:
                jump_host = remote_connect_info.jump.server.host_name
            forward_tunnel_info = SSHTunnelInfo(
                ssh_client=ssh_tunnel_client,
                service_group=service_group,
                service_name=service_name,
                jump_host=jump_host,
                remote_host=remote_host,
                remote_port=int(remote_port),
                local_port=int(local_port),
                dest_host=dest_host,
                dest_port=dest_port,
                cookie=str(service_data["cookie"]),
                thread_process=thread_process,
                forward_server=forward_server,
                forward_tunnel=True,
            )
            lock = threading.Lock()
            with lock:
                g_open_tunnel_list.append(forward_tunnel_info)
            if service_data["cookie"] is not None:
                self.send_response(
                    "port:" + str(local_port) + ",cookie:" + str(service_data["cookie"]) + ",msg:no_error"
                )
            else:
                self.send_response("port:" + str(local_port) + ",msg:no_error")
        else:
            self.send_response("port:-1,msg:cannot_start_forward_server")
        return

    def check_forward_tunnel_exist(self, jump_host, remote_host, dest_host, dest_port):
        for tnnl in g_open_tunnel_list:
            if (
                (tnnl.forward_tunnel is True)
                and (jump_host == tnnl.jump_host)
                and (remote_host == tnnl.remote_host)
                and (dest_host == tnnl.dest_host)
                and (dest_port == tnnl.dest_port)
            ):
                return tnnl.local_port
        return None

    def get_jump_remote_from_config(self, req_qry):
        if "version" not in g_server_config_data:
            verbose("No config version, assuming v0.1")
            return self.get_jump_remote_from_config_v01(req_qry)
        if g_server_config_data["version"] == "v0.1":
            verbose("Config version v0.1")
            return self.get_jump_remote_from_config_v01(req_qry)
        verbose("Unsupported configuration version: %s" % (g_server_config_data["version"],))
        return None

    def get_jump_remote_from_config_v01(self, req_qry):
        req_group_name = req_qry["group"][0]
        req_service_name = req_qry["service"][0]
        for group_name in g_server_config_data:
            if group_name == req_group_name:
                service_group = g_server_config_data[group_name]
                for service_name in service_group:
                    if service_name == req_service_name:
                        service_data = service_group[service_name]
                        if (
                            ("host" not in service_data)
                            or ("protocol" not in service_data)
                            or ("authentication" not in service_data)
                        ):
                            return None
                        req_def = {}
                        req_def["service_group"] = req_group_name
                        req_def["service_name"] = service_name
                        req_def["username"] = None
                        if "user" in service_data:
                            req_def["username"] = service_data["user"]
                        req_def["remote_host"] = service_data["host"]
                        req_def["jumphost"] = None
                        if "jumphost" in service_data:
                            req_def["jumphost"] = service_data["jumphost"]
                        req_def["jumpuser"] = None
                        if "jumpuser" in service_data:
                            req_def["jumpuser"] = service_data["jumpuser"]
                        req_def["tunneltype"] = "permanent"
                        if "tunneltype" in service_data:
                            req_def["tunneltype"] = service_data["tunneltype"]
                        req_def["conn"] = service_data["protocol"]
                        req_def["auth"] = service_data["authentication"]
                        return req_def
        return None

    def do_connect_port(self, req_qry):
        global g_open_tunnel_list
        if "dhost" not in req_qry:
            self.send_response("port:-1,msg:missing_destination_host_in_request")
            return
        if "dport" not in req_qry:
            self.send_response("port:-1,msg:missing_destination_port_in_request")
            return
        req_def = {}
        if ("group" in req_qry) or ("service" in req_qry):
            if "service" not in req_qry:
                self.send_response("port:-1,msg:missing_service_in_request")
                return
            if "group" not in req_qry:
                self.send_response("port:-1,msg:missing_group_in_request")
                return
            req_def = self.get_jump_remote_from_config(req_qry)
            if req_def is None:
                self.send_response("port:-1,msg:no_matching_group_or_service_in_config_file")
                return
        else:
            if "rhost" not in req_qry:
                self.send_response("port:-1,msg:missing_remote_host_in_request")
                return
            req_def["remote_host"] = req_qry["rhost"][0]
            req_def["username"] = None
            if "ruser" in req_qry:
                req_def["username"] = req_qry["ruser"][0]
            req_def["jumphost"] = None
            req_def["jumpuser"] = None
            if "jhost" in req_qry:
                req_def["jumphost"] = req_qry["jhost"][0]
                if "juser" in req_qry:
                    req_def["jumpuser"] = req_qry["juser"][0]
            req_def["auth"] = "Password/Passcode"
        req_def["dest_host"] = req_qry["dhost"][0]
        req_def["dest_port"] = req_qry["dport"][0]

        connected_dest_port = self.check_forward_tunnel_exist(
            req_def["jumphost"],
            req_def["remote_host"],
            req_def["dest_host"],
            int(req_def["dest_port"]),
        )
        if connected_dest_port is not None:
            self.send_response("port:" + str(connected_dest_port) + ",msg:no_error")
            return

        ssh_connect_remote = self.login_connect_remote(req_def)
        if ssh_connect_remote is None:
            self.send_response("port:-1,msg:cannot_connect_to_remote_host_port")
            return

        ssh_tunnel_client = SSHLocalRemoteTunnel(ssh_connect_remote)

        remote_host = req_def["remote_host"]
        remote_port = SSH_PORT
        dest_host = req_def["dest_host"]
        dest_port = int(req_def["dest_port"])
        remote_srvr = SSHServerInfo(dest_host, dest_port)

        local_port = self.get_local_port(requested_port=None)
        if local_port is None:
            self.send_response("port:-1,msg:cannot_find_local_port")
            return

        temptunnel = False
        if "transient" in req_def["tunneltype"]:
            temptunnel = True

        forward_server, thread_process, error_no = ssh_tunnel_client.start_forward_server(
            local_port, remote_srvr, temptunnel=temptunnel
        )
        if error_no == SSH_NO_ERROR:
            remote_connect_info = ssh_tunnel_client.get_remote_connection_info()
            jump_host = None
            if remote_connect_info.jump is not None:
                jump_host = remote_connect_info.jump.server.host_name
            forward_tunnel_info = SSHTunnelInfo(
                ssh_client=ssh_tunnel_client,
                jump_host=jump_host,
                remote_host=remote_host,
                remote_port=int(remote_port),
                local_port=int(local_port),
                dest_host=dest_host,
                dest_port=dest_port,
                thread_process=thread_process,
                forward_server=forward_server,
                forward_tunnel=True,
            )
            lock = threading.Lock()
            with lock:
                g_open_tunnel_list.append(forward_tunnel_info)
            self.send_response("port:" + str(local_port) + ",msg:no_error")
        else:
            self.send_response("port:-1,msg:cannot_start_forward_server")
        return

    def do_reverse_port_forward(self, req_qry):
        print("Reverse port forwarding request: ", req_qry)
        if "group" not in req_qry:
            self.send_response("port:-1,msg:missing_group_in_request")
            return
        if "service" not in req_qry:
            self.send_response("port:-1,msg:missing_service_in_request")
            return
        if "dest_host" not in req_qry:
            self.send_response("port:-1,msg:missing_dest_host_in_request")
            return
        if "remote_port" not in req_qry:
            self.send_response("port:-1,msg:missing_remote_port_in_request")
            return
        if "dest_port" not in req_qry:
            self.send_response("port:-1,msg:missing_dest_port_in_request")
            return

        req_def = self.parse_service_request(req_qry)
        if req_def is None:
            self.send_response("port:-1,msg:incomplete_service_definition")
            return
        req_def["dest_host"] = req_qry["dest_host"][0]
        req_def["dest_port"] = req_qry["dest_port"][0]
        req_def["remote_port"] = req_qry["remote_port"][0]

        print("Reverse port request: ", req_def)
        self.execute_reverse_port_forwarding(req_def)
        return

    def execute_reverse_port_forwarding(self, req_def):
        global g_open_tunnel_list
        print("Request :", req_def)

        ssh_connect_remote = self.login_connect_remote(req_def)
        if ssh_connect_remote is None:
            self.send_response("port:-1,msg:cannot_connect_to_remote_host_port")
            return

        remote_port = int(req_def["remote_port"])
        dest_host = req_def["dest_host"]
        dest_port = int(req_def["dest_port"])
        ssh_tunnel_client = SSHLocalRemoteTunnel(ssh_connect_remote)

        thread_process = ssh_tunnel_client.reverse_forward_tunnel(
            remote_server_port=remote_port, dest_host=dest_host, dest_port=dest_port
        )
        if thread_process is None:
            self.send_response("port:-1,msg:cannot_connect_to_remote_host_port")
            return

        reverse_tunnel_info = SSHTunnelInfo(
            ssh_client=ssh_tunnel_client,
            remote_host=req_def["remote_host"],
            remote_port=remote_port,
            dest_host=dest_host,
            dest_port=dest_port,
            thread_process=thread_process,
            reverse_tunnel=True,
        )
        lock = threading.Lock()
        with lock:
            g_open_tunnel_list.append(reverse_tunnel_info)
        self.send_response("port:" + str(reverse_tunnel_info["remote_port"]) + "msg:no_error")

        return

    def send_key(self, req_qry):
        if "id" not in req_qry:
            self.send_response("key:0,msg:missing_key_id_in_request")
            return

        key_id = req_qry["id"][0]
        if key_id in g_keys:
            self.send_response("key:" + g_keys[key_id].hex() + ",msg:no_error")
        else:
            self.send_response("key:0,msg:cannot_find_key_id")


def parse_arguments(argv):
    server_port = PORT
    config_file = REMOTE_CONFIG_FILE
    proxy_command = None
    try:
        opts, args = getopt.getopt(argv, "hc:p:x:")
    except getopt.GetoptError:
        print("python ssh_tunnel_server.py -c <config file> -p <server port> -x <proxy command>")
        sys.exit(2)
    for opt, arg in opts:
        if opt == "-h":
            print("python ssh_tunnel_server.py -c <config file> -p <server port> -x <proxy command>")
            sys.exit(0)
        elif opt == "-c":
            config_file = arg
        elif opt == "-p":
            server_port = int(arg)
        elif opt == "-x":
            proxy_command = arg
    return config_file, server_port, proxy_command


def removeTunnel(dest_host, dest_port):
    global g_open_tunnel_list
    for tnnl in g_open_tunnel_list:
        if (tnnl.forward_tunnel is True) and (dest_host == tnnl.dest_host) and (dest_port == tnnl.dest_port):
            print("Shutting down: ", dest_host, " port: ", dest_port)
            tnnl.forward_server.shutdown()
            tnnl.forward_server.server_close()
            g_open_tunnel_list.remove(tnnl)


def run_queue():
    while True:
        (dest_host, dest_port) = g_queue.get()
        print("QUEUE :", dest_host, " port: ", dest_port)
        lock = threading.Lock()
        with lock:
            removeTunnel(dest_host, dest_port)


g_keys: dict = {}


def read_keys():
    keys_pattern = os.path.expanduser("~/.config/hpc-campaign/keys/*")
    keyFileList = glob.glob(keys_pattern)
    for f in keyFileList:
        print(f"Loading key {f}")
        key = Key()
        key.read(f)
        print(f"  created on: {datetime.fromisoformat(key.date)}")
        print(f"        note: {key.note}")
        print(f"        uuid: {key.id}")
        if key.salt:
            print("      encryption: password")
        else:
            print("      encryption: none")
        try:
            g_keys[key.id] = key.get_decrypted_key()
        except Exception:
            print(f"Decryption of key failed. Ignoring key {f}")
    print(g_keys)


class ReuseAddrTCPServer(SocketServer.TCPServer):
    allow_reuse_address = True


def start_server(argv):
    global g_remote_conn_list
    global g_open_tunnel_list
    global g_server_config_data
    global g_proxy_command
    g_remote_conn_list = []
    g_open_tunnel_list = []
    config_file, server_port, g_proxy_command = parse_arguments(argv)
    if os.path.isfile(config_file) is True:
        g_server_config_data = read_yaml_config(config_file)
    else:
        g_server_config_data = {}

    read_keys()

    thread = threading.Thread(target=run_queue, args=())
    thread.daemon = True
    thread.start()
    while thread.is_alive() is False:
        time.sleep(THREAD_CHECK_TIME)

    #    SocketServer.TCPServer.allow_reuse_address = True
    #    with SocketServer.TCPServer((HOST, server_port),MyTCPHandler) as server:
    with ReuseAddrTCPServer((HOST, server_port), MyTCPHandler) as server:
        print("SSH Tunnel Server: ", HOST, server_port)
        if g_proxy_command:
            print("Using proxy command", g_proxy_command)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print("Closing")
            server.server_close()
        except Exception as e:
            print("Exception: ", e)
            server.server_close()


def main(args=sys.argv[1:], prog=None):
    start_server(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import yaml
from os.path import expanduser

ACA_VERSION = "0.6"
# 0.2 added key encryption (added table key, modfified table bpdataset)
# 0.3 generate UUID for each bpdataset (modified table bpdataset)
# 0.4 added h5dataset table
# 0.5 reorganized to "dataset" table, plus text and images, replicas, archives
# 0.6 redefines "archive" as TAR files, throws away 0.5's archive concept

REDIS_PORT = 6379


class Config:
    """User config and Hosts config file parser"""

    def __init__(self):
        self.campaign_store_path = None
        self.host_name = None
        self.cache_path = None
        self.verbose: int = 0
        path = expanduser("~/.config/hpc-campaign/config.yaml")
        try:
            doc = {}
            with open(path) as f:
                doc = yaml.safe_load(f)
            camp = doc.get("Campaign")
            if isinstance(camp, dict):
                for key, value in camp.items():
                    if key == "campaignstorepath":
                        self.campaign_store_path = expanduser(value)
                    if key == "hostname":
                        self.host_name = value
                    if key == "cachepath":
                        self.cache_path = value
                    if key == "verbose":
                        self.verbose = value
        except FileNotFoundError:
            None

    def read_host_config(self) -> dict:
        path = expanduser("~/.config/hpc-campaign/hosts.yaml")
        doc = {}
        try:
            with open(path) as f:
                doc = yaml.safe_load(f)
        except FileNotFoundError:
            None
        return doc



#!/usr/bin/env python3
import argparse
from os import remove
from os.path import exists
from .key import Key


def setup_args(args=None, prog=None):
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument(
        "command",
        help="Command: generate/verify/info",
        choices=["generate", "verify", "info"],
    )
    parser.add_argument("path", help="path of keyfile")
    parser.add_argument("--verbose", "-v", help="More verbosity", action="count", default=0)
    parser.add_argument("--password", "-p", help="Protect with password", action="store_true", default=False)
    args = parser.parse_args(args=args)

    if args.verbose > 0:
        print(f"# Verbosity = {args.verbose}")
        print(f"# Command = {args.command}")
        print(f"# Path = {args.path}")
        print(f"# Password required = {args.password}")
    return args


def check_path_for_creation(path: str):
    if exists(path):
        print(f"The file {path} already exist. " "Do not destroy a key that is in use of existing campaign archives. ")
        while True:
            print("Do you want to overwrite Y/N? ", end="")
            answer = input()
            if answer == "N" or answer == "n":
                exit(1)
            if answer == "Y" or answer == "y":
                break
    else:
        try:
            with open(path, "wb") as f:
                f.write(b"test")
            remove(path)
        except Exception:
            print(f"Could not create/write to {path}")
            exit(1)


def check_path_for_reading(path: str):
    if not exists(path):
        print(f"Could not find {path}")
        exit(1)


def main(args=None, prog=None):
    args = setup_args(args=args, prog=prog)
    key = Key()
    if args.command == "generate":
        check_path_for_creation(args.path)
        key.generate_interactive(args.password)
        key.write(args.path)

    elif args.command == "verify":
        check_path_for_reading(args.path)
        key.read(args.path)
        key.info(True)

    elif args.command == "info":
        check_path_for_reading(args.path)
        key.read(args.path)
        key.info(False)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import glob
import sqlite3
import zlib
import uuid
import nacl.secret
import nacl.utils
import csv
from dateutil.parser import parse
from hashlib import sha1
from os import chdir, getcwd, remove, stat
from os.path import exists, isdir, basename, join
from pathlib import Path
from PIL import Image
from re import sub
from socket import getfqdn
from time import time_ns, sleep

from .key import read_key
from .config import ACA_VERSION
from .utils import (
    timestamp_to_str,
    SQLCommit,
    SQLExecute,
    SQLErrorList,
    get_folder_size,
    sizeof_fmt,
)
from .upgrade import UpgradeACA
from .taridx import TARTYPES
from .hdf5_metadata import copy_hdf5_file_without_data, IsHDF5Dataset
from .manager_args import ArgParser

CURRENT_TIME = time_ns()


def CheckCampaignStore(args):
    if args.campaign_store is not None and not isdir(args.campaign_store):
        print("ERROR: Campaign directory " + args.campaign_store + " does not exist", flush=True)
        exit(1)


def CheckLocalCampaignDir(args):
    if not isdir(args.LocalCampaignDir):
        print(
            "ERROR: Shot campaign data '"
            + args.LocalCampaignDir
            + "' does not exist. Run this command where the code was executed.",
            flush=True,
        )
        exit(1)


def parse_date_to_utc(date, fmt=None):
    if fmt is None:
        fmt = "%Y-%m-%d %H:%M:%S %z"  # Defaults to : 2022-08-31 07:47:30 -0000
    get_date_obj = parse(str(date))
    return get_date_obj.timestamp()


def IsADIOSDataset(dataset):
    if not isdir(dataset):
        return False
    if not exists(dataset + "/" + "md.idx"):
        return False
    if not exists(dataset + "/" + "data.0"):
        return False
    return True


def compressBytes(b: bytes) -> tuple[bytes, int, int, str]:
    compObj = zlib.compressobj()
    compressed = bytearray()
    len_orig = len(b)
    len_compressed = 0
    checksum = sha1(b)

    cBlock = compObj.compress(b)
    compressed += cBlock
    len_compressed += len(cBlock)

    cBlock = compObj.flush()
    compressed += cBlock
    len_compressed += len(cBlock)

    return bytes(memoryview(compressed)), len_orig, len_compressed, checksum.hexdigest()


def compressFile(f) -> tuple[bytes, int, int, str]:
    compObj = zlib.compressobj()
    compressed = bytearray()
    blocksize = 1073741824  # 1GB #1024*1048576
    len_orig = 0
    len_compressed = 0
    checksum = sha1()
    block = f.read(blocksize)
    while block:
        len_orig += len(block)
        cBlock = compObj.compress(block)
        compressed += cBlock
        len_compressed += len(cBlock)
        checksum.update(block)
        block = f.read(blocksize)
    cBlock = compObj.flush()
    compressed += cBlock
    len_compressed += len(cBlock)

    return bytes(memoryview(compressed)), len_orig, len_compressed, checksum.hexdigest()


def decompressBuffer(buf: bytearray):
    data = zlib.decompress(buf)
    return data


def encryptBuffer(args: argparse.Namespace, buf: bytes):
    if args.encryption_key:
        box = nacl.secret.SecretBox(args.encryption_key)
        nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)
        e = box.encrypt(buf, nonce)
        print("Encoded buffer size: ", len(e))
        return e
    else:
        return buf


def lastrowid_or_zero(curDS: sqlite3.Cursor) -> int:
    rowID = curDS.lastrowid
    if not rowID:
        rowID = 0
    return rowID


def AddFileToArchive(
    args: argparse.Namespace,
    filename: str,
    cur: sqlite3.Cursor,
    repID: int,
    mt: float = 0.0,
    filename_as_recorded: str = "",
    compress: bool = True,
    content: bytes = bytes(),
    indent: str = "",
):
    if compress:
        compressed = 1
        if content:
            compressed_data, len_orig, len_compressed, checksum = compressBytes(content)
        else:
            try:
                with open(filename, "rb") as f:
                    compressed_data, len_orig, len_compressed, checksum = compressFile(f)

            except IOError:
                print(f"{indent}ERROR While reading file {filename}")
                return
    else:
        compressed = 0
        if content:
            compressed_data = content
        else:
            try:
                with open(filename, "rb") as f:
                    compressed_data = f.read()
            except IOError:
                print(f"{indent}ERROR While reading file {filename}")
                return
        len_orig = len(compressed_data)
        len_compressed = len_orig
        checksum = sha1(compressed_data).hexdigest()

    encrypted_data = encryptBuffer(args, compressed_data)

    if mt == 0.0:
        statres = stat(filename)
        mt = statres.st_mtime_ns

    if len(filename_as_recorded) == 0:
        filename_as_recorded = filename

    SQLExecute(
        cur,
        "insert into file "
        "(replicaid, name, compression, lenorig, lencompressed, modtime, checksum, data) "
        "values (?, ?, ?, ?, ?, ?, ?, ?) "
        "on conflict (replicaid, name) do update "
        "set compression = ?, lenorig = ?, lencompressed = ?, modtime = ?, checksum = ?, data = ?",
        (
            repID,
            filename_as_recorded,
            compressed,
            len_orig,
            len_compressed,
            mt,
            checksum,
            encrypted_data,
            compressed,
            len_orig,
            len_compressed,
            mt,
            checksum,
            encrypted_data,
        ),
    )


def AddReplicaToArchive(
    args: argparse.Namespace,
    hostID: int,
    dirID: int,
    archiveID: int,
    keyID: int,
    dataset: str,
    cur: sqlite3.Cursor,
    datasetid: int,
    mt: float,
    size: int,
    indent: str = "",
) -> int:

    print(f"{indent}Add replica {dataset} to archive")
    print(
        f"{indent}AddReplicaToArchive(host={hostID}, dir={dirID}, archive={archiveID}, key={keyID}, name={dataset}"
        f" dsid={datasetid}, time={mt}, size={size})"
    )
    curDS = SQLExecute(
        cur,
        "insert into replica (datasetid, hostid, dirid, archiveid, name, modtime, deltime, keyid, size) "
        "values  (?, ?, ?, ?, ?, ?, ?, ?, ?) "
        "on conflict (datasetid, hostid, dirid, archiveid, name) "
        "do update set modtime = ?, deltime = ?, keyid = ?, size = ? "
        "returning rowid",
        (datasetid, hostID, dirID, archiveID, dataset, mt, 0, keyID, size, mt, 0, keyID, size),
    )
    rowID = curDS.fetchone()[0]
    print(f"{indent}  Replica rowid = {rowID}")
    return rowID


def AddDatasetToArchive(
    args: argparse.Namespace, name: str, cur: sqlite3.Cursor, uniqueID: str, format: str, mt: float, indent: str = ""
) -> int:

    print(f"{indent}Add dataset {name} to archive")
    curDS = SQLExecute(
        cur,
        "insert into dataset (name, uuid, modtime, deltime, fileformat, tsid, tsorder) "
        "values  (?, ?, ?, ?, ?, ?, ?) "
        "on conflict (name) do update set deltime = ? "
        "returning rowid",
        (name, uniqueID, mt, 0, format, 0, 0, 0),
    )
    datasetID = curDS.fetchone()[0]
    return datasetID


def AddResolutionToArchive(
    args: argparse.Namespace, repID: int, x: int, y: int, cur: sqlite3.Cursor, indent: str = ""
) -> int:

    print(f"{indent}Add resolution {x} {y} for replica {repID} to archive")
    curDS = SQLExecute(
        cur,
        "insert into resolution (replicaid, x, y) "
        "values  (?, ?, ?) "
        "on conflict (replicaid) do update set x = ?, y = ? returning rowid",
        (repID, x, y, x, y),
    )
    rowID = curDS.fetchone()[0]
    return rowID


def ProcessDatasets(
    args: argparse.Namespace,
    cur: sqlite3.Cursor,
    hostID: int,
    dirID: int,
    keyID: int,
    dirpath: str,
    location: str,
):
    for entry in args.files:
        dataset = entry
        if args.name is not None:
            dataset = args.name
        uniqueID = uuid.uuid3(uuid.NAMESPACE_URL, location + "/" + entry).hex
        dsID = 0

        if args.remote_data:
            filesize = 0
            if args.s3_datetime:
                mt = parse_date_to_utc(args.s3_datetime)
            else:
                mt = 0
        else:
            statres = stat(entry)
            mt = statres.st_mtime_ns
            filesize = statres.st_size

        if args.remote_data:
            dsID = AddDatasetToArchive(args, dataset, cur, uniqueID, "ADIOS", mt)
            repID = AddReplicaToArchive(args, hostID, dirID, 0, keyID, entry, cur, dsID, mt, filesize, indent="  ")
        elif IsADIOSDataset(entry):
            dsID = AddDatasetToArchive(args, dataset, cur, uniqueID, "ADIOS", mt)
            filesize = get_folder_size(entry)
            repID = AddReplicaToArchive(args, hostID, dirID, 0, keyID, entry, cur, dsID, mt, filesize, indent="  ")
            cwd = getcwd()
            chdir(entry)
            mdFileList = glob.glob("*md.*")
            profileList = glob.glob("profiling.json")
            files = mdFileList + profileList
            for f in files:
                AddFileToArchive(args, f, cur, repID)
            chdir(cwd)
        elif IsHDF5Dataset(entry):
            mdfilename = "/tmp/md_" + basename(entry)
            copy_hdf5_file_without_data(entry, mdfilename)
            dsID = AddDatasetToArchive(args, dataset, cur, uniqueID, "HDF5", mt)
            repID = AddReplicaToArchive(args, hostID, dirID, 0, keyID, entry, cur, dsID, mt, filesize, indent="  ")
            AddFileToArchive(args, mdfilename, cur, repID, mt, basename(entry))
            remove(mdfilename)
        else:
            print(f"WARNING: Dataset {dataset} is neither an ADIOS nor an HDF5 dataset. Skip")


def ProcessTextFiles(
    args: argparse.Namespace,
    cur: sqlite3.Cursor,
    hostID: int,
    dirID: int,
    keyID: int,
    dirpath: str,
    location: str,
):
    for entry in args.files:
        print(f"Process entry {entry}:")
        dataset = entry
        if args.name is not None:
            dataset = args.name
        statres = stat(entry)
        ct = statres.st_mtime_ns
        filesize = statres.st_size
        uniqueID = uuid.uuid3(uuid.NAMESPACE_URL, location + "/" + entry).hex
        dsID = AddDatasetToArchive(args, dataset, cur, uniqueID, "TEXT", ct)
        repID = AddReplicaToArchive(args, hostID, dirID, 0, keyID, entry, cur, dsID, ct, filesize, indent="  ")
        if args.store:
            AddFileToArchive(args, entry, cur, repID, ct, basename(entry))


def ProcessImage(
    args: argparse.Namespace,
    cur: sqlite3.Cursor,
    hostID: int,
    dirID: int,
    keyID: int,
    dirpath: str,
    location: str,
):
    dataset = args.file
    if args.name is not None:
        dataset = args.name

    statres = stat(args.file)
    mt = statres.st_mtime_ns
    filesize = statres.st_size
    uniqueID = uuid.uuid3(uuid.NAMESPACE_URL, location + "/" + args.file).hex
    print(f"Process image {location}/{args.file}")

    img = Image.open(args.file)
    imgres = img.size

    dsID = AddDatasetToArchive(args, dataset, cur, uniqueID, "IMAGE", mt, indent="  ")
    repID = AddReplicaToArchive(args, hostID, dirID, 0, keyID, args.file, cur, dsID, mt, filesize, indent="  ")
    AddResolutionToArchive(args, repID, imgres[0], imgres[1], cur, indent="  ")

    if args.store or args.thumbnail is not None:
        imgsuffix = Path(args.file).suffix
        if args.store:
            print("Storing the image in the archive")
            resname = f"{imgres[0]}x{imgres[1]}{imgsuffix}"
            AddFileToArchive(args, args.file, cur, repID, mt, resname, compress=False, indent="  ")

        else:
            print(f"  Make thumbnail image with resolution {args.thumbnail}")
            img.thumbnail(args.thumbnail)
            imgres = img.size
            resname = f"{imgres[0]}x{imgres[1]}{imgsuffix}"
            now = time_ns()
            thumbfilename = "/tmp/" + basename(resname)
            img.save(thumbfilename)
            statres = stat(thumbfilename)
            mt = statres.st_mtime_ns
            filesize = statres.st_size
            thumbrepID = AddReplicaToArchive(
                args, hostID, dirID, 0, keyID, join("thumbnails", args.file), cur, dsID, now, filesize, indent="  "
            )
            AddFileToArchive(args, thumbfilename, cur, thumbrepID, now, resname, compress=False, indent="  ")
            AddResolutionToArchive(args, thumbrepID, imgres[0], imgres[1], cur, indent="  ")
            remove(thumbfilename)


def ArchiveDataset(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection, indent: str = ""):
    # Find dataset
    res = SQLExecute(cur, f'select rowid, fileformat from dataset where name = "{args.name}"')
    rows = res.fetchall()
    if len(rows) == 0:
        raise Exception(f"Dataset not found: {args.name} ")

    datasetid: int = rows[0][0]
    format: str = rows[0][1]

    # Find archive dir
    res = SQLExecute(cur, f"select hostid, name from directory where rowid = {args.dirid}")
    rows = res.fetchall()
    if len(rows) == 0:
        raise Exception(f"Directory ID not found: {args.dirid} ")

    hostID: int = rows[0][0]
    dir_name: str = rows[0][1]

    if args.archiveid is None:
        res = SQLExecute(cur, f"select rowid from archive where dirid = {args.dirid}")
        rows = res.fetchall()
        if len(rows) == 0:
            raise Exception(f"Directory {dir_name} with ID {args.dirid} is not an archival storage directory")
        archiveID = rows[0][0]
    else:
        res = SQLExecute(cur, f"select rowid, dirid from archive where rowid = {args.archiveid}")
        rows = res.fetchall()
        if len(rows) == 0:
            raise Exception(f"Archive ID {args.archiveid} is not found in the archive list")
        archiveID = args.archiveid
        dirID = rows[0][1]
        if dirID != args.dirid:
            raise Exception(f"Archive ID {args.archiveid} belongs to dir ID {dirID}, not to {args.dirid}")

    # Check replicas of dataset and see if there is conflict (need --replica option)
    orig_repID: int = args.replica
    if args.replica is None:
        res = SQLExecute(cur, f"select rowid, archiveid, deltime from replica where datasetid = {datasetid}")
        rows = res.fetchall()
        delrows = []
        live_nonarch_rows = []
        live_arch_rows = []
        for row in rows:
            if row[2] == 0:
                if row[1] == 0:
                    live_nonarch_rows.append(row)
                else:
                    live_arch_rows.append(row)
            else:
                delrows.append(row)
        if len(live_nonarch_rows) > 1:
            raise Exception(
                f"There are {len(live_nonarch_rows)} non-deleted, not-in-archive, replicas for this dataset. "
                f"Use --replica to identify which is archived now. Replicas: {[r[0] for r in live_nonarch_rows]}"
            )
        elif len(live_nonarch_rows) + len(live_arch_rows) == 0:
            if format == "ADIOS" or format == "HDF5":
                raise Exception(
                    f"There are no replicas for a {format} dataset. Cannot archive without "
                    "access to the embedded metadata files of a replica"
                )
            if len(delrows) == 1:
                orig_repID = delrows[0][0]
            else:
                raise Exception(
                    f"There are no replicas but {len(delrows)} deleted replicas for this {format} dataset. "
                    "Use --replica to identify which deleted replica is archived."
                    f"Deleted replicas: {[r[0] for r in delrows]}"
                )
        else:
            if len(live_nonarch_rows) > 0:
                orig_repID = live_nonarch_rows[0][0]
            elif len(live_arch_rows) > 1:
                raise Exception(
                    f"There are {len(live_arch_rows)} archived replicas for this dataset. "
                    f"Use --replica to identify which is archived now. Replicas: {[r[0] for r in live_arch_rows]}"
                )
            else:
                orig_repID = live_arch_rows[0][0]

    # get name and KeyID for selected replica
    res = SQLExecute(cur, f"select datasetid, name, modtime, keyid, size from replica where rowid = {orig_repID}")
    row = res.fetchone()
    if datasetid != row[0]:
        res = SQLExecute(cur, f'select name from dataset where rowid = "{row[0]}"')
        wrong_dsname = res.fetchone()[0]
        raise Exception(f"Replica belongs to dataset {wrong_dsname}, not this dataset")
    replicaName: str = row[1]
    mt: int = row[2]
    keyID: int = row[3]
    filesize: int = row[4]

    # Create new replica for this dataset
    dsname = replicaName
    if args.newpath:
        dsname = args.newpath

    repID = AddReplicaToArchive(
        args, hostID, args.dirid, archiveID, keyID, dsname, cur, datasetid, mt, filesize, indent=indent
    )

    # if replica has Resolution, copy that to new replica
    res = SQLExecute(cur, f"select x, y from resolution where replicaid = {orig_repID}")
    rows = res.fetchall()
    if len(rows) > 0:
        x = rows[0][0]
        y = rows[0][1]
        AddResolutionToArchive(args, repID, x, y, cur, indent=indent)

    # # if replica has Accuracy, copy that to new replica
    # res = SQLExecute(cur, f"select accuracy, norm, relative from accuracy where replicaid = {orig_repID}")
    # rows = res.fetchall()
    # if len(rows) > 0:
    #     accuracy = rows[0][0]
    #     norm = rows[0][1]
    #     relative = rows[0][2]
    #     AddAccuracyToArchive(args, repID, accuracy, norm, relative, cur)

    # if --move, delete the original replica but assign embedded files to archived replica
    # otherwise, make a copy of all embedded files
    if args.move:
        SQLExecute(cur, f"update file set replicaid = {repID} where replicaid = {orig_repID}")
        DeleteReplica(args, cur, con, orig_repID, False, indent=indent)
    else:
        res = SQLExecute(
            cur,
            "select name, compression, lenorig, lencompressed, modtime, checksum, data "
            f"from file where replicaid = {orig_repID}",
        )
        files = res.fetchall()
        print(f"{indent}Copying {len(files)} files from original replica to archived one")
        for f in files:
            SQLExecute(
                cur,
                "insert into file values (?, ?, ?, ?, ?, ?, ?, ?) "
                "on conflict (replicaid, name) do update set "
                "compression = ?, lenorig = ?, lencompressed = ?, modtime = ?, checksum = ?, data = ?",
                (repID, f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[1], f[2], f[3], f[4], f[5], f[6]),
            )

    SQLCommit(con)
    return repID


def AddTimeSeries(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    if args.remove:
        res = SQLExecute(cur, f'select tsid from timeseries where name = "{args.name}"')
        rows = res.fetchall()
        if len(rows) > 0:
            tsID = rows[-1][0]
            print(f"Remove {args.name} from time-series but leave datasets alone")
            res = SQLExecute(cur, f'delete from timeseries where name = "{args.name}"')
            curDS = SQLExecute(cur, f'update dataset set tsid = 0, tsorder = 0 where tsid = "{tsID}"')
        else:
            print(f"Time series {args.name} was not found")
        SQLCommit(con)
        return

    print(f"Add {args.name} to time-series")
    # we need to know if it already exists
    ts_exists = False
    res = SQLExecute(cur, f'select tsid from timeseries where name = "{args.name}"')
    rows = res.fetchall()
    if len(rows) > 0:
        ts_exists = True

    # insert/update timeseries
    curTS = SQLExecute(
        cur,
        "insert into timeseries (name) values  (?) " "on conflict (name) do update set name = ? returning rowid",
        (args.name, args.name),
    )
    tsID = curTS.fetchone()[0]
    print(f"Time series ID = {tsID}, already existed = {ts_exists}")

    # if --replace, "delete" the existing dataset connections
    tsorder = 0
    if args.replace:
        curDS = SQLExecute(cur, f'update dataset set tsid = 0, tsorder = 0 where tsid = "{tsID}"')
    else:
        # otherwise we need to know how many datasets we have already
        res = SQLExecute(cur, f"select tsorder from dataset where tsid = {tsID} order by tsorder")
        rows = res.fetchall()
        if len(rows) > 0:
            tsorder = rows[-1][0] + 1

    for dsname in args.dataset:
        curDS = SQLExecute(
            cur,
            f"update dataset set tsid = {tsID}, tsorder = {tsorder} "
            + f'where name = "{dsname}" returning rowid, name',
        )
        ret = curDS.fetchone()
        if ret is None:
            print(f"    {dsname}  Error: dataset is not in the database, skipping")
        else:
            rowID = ret[0]
            name = ret[1]
            print(f"    {name} (dataset {rowID}) tsorder = {tsorder}")
            tsorder += 1

    SQLCommit(con)


def GetHostName(args: argparse.Namespace):
    if args.s3_endpoint:
        longhost = args.s3_endpoint
    else:
        longhost = getfqdn()
        if longhost.startswith("login"):
            longhost = sub("^login[0-9]*\\.", "", longhost)
        if longhost.startswith("batch"):
            longhost = sub("^batch[0-9]*\\.", "", longhost)

    if args.hostname is None:
        shorthost = longhost.split(".")[0]
    else:
        shorthost = args.hostname
    return longhost, shorthost


def AddHostName(longHostName, shortHostName, cur: sqlite3.Cursor, default_protocol: str = "", indent: str = "") -> int:
    res = SQLExecute(cur, 'select rowid from host where hostname = "' + shortHostName + '"')
    row = res.fetchone()
    if row is not None:
        hostID = row[0]
        print(f"{indent}Found host {shortHostName} in database, rowid = {hostID}")
    else:
        curHost = SQLExecute(
            cur,
            "insert into host values (?, ?, ?, ?, ?)",
            (shortHostName, longHostName, CURRENT_TIME, 0, default_protocol),
        )
        hostID = lastrowid_or_zero(curHost)
        print(f"{indent}Inserted host {shortHostName} into database, rowid = {hostID}, longhostname = {longHostName}")
    return hostID


def AddDirectory(hostID: int, path: str, cur: sqlite3.Cursor, indent: str = "") -> int:
    res = SQLExecute(
        cur,
        "select rowid from directory where hostid = " + str(hostID) + ' and name = "' + path + '"',
    )
    row = res.fetchone()
    if row is not None:
        dirID = row[0]
        print(f"{indent}Found directory {path} with hostID {hostID} in database, rowid = {dirID}")
    else:
        curDirectory = SQLExecute(cur, "insert into directory values (?, ?, ?, ?)", (hostID, path, CURRENT_TIME, 0))
        dirID = lastrowid_or_zero(curDirectory)
        print(f"{indent}Inserted directory {path} into database, rowid = {dirID}")
    return dirID


def AddKeyID(key_id: str, cur: sqlite3.Cursor) -> int:
    if key_id:
        res = SQLExecute(cur, 'select rowid from key where keyid = "' + key_id + '"')
        row = res.fetchone()
        if row is not None:
            keyID = row[0]
            print(f"Found key {key_id} in database, rowid = {keyID}")
        else:
            cmd = f'insert into key values ("{(key_id)}")'
            curKey = SQLExecute(cur, cmd)
            # curKey = SQLExecute(cur,"insert into key values (?)", (key_id))
            keyID = lastrowid_or_zero(curKey)
            print(f"Inserted key {key_id} into database, rowid = {keyID}")
        return keyID
    else:
        return 0  # an invalid row id


def ArchiveIdxReplica(
    dsname: str,
    dirID: int,
    archiveID: int,
    replicaID: int,
    entries: dict[str, list[int]],
    cur: sqlite3.Cursor,
    con: sqlite3.Connection,
    indent: str = "",
):
    # Archive replica
    args = argparse.Namespace()
    args.name = dsname
    args.dirid = dirID
    args.archiveid = archiveID
    args.replica = replicaID
    args.move = False
    args.newpath = ""

    archived_replicaID = ArchiveDataset(args, cur, con, indent=indent + "  ")
    if archived_replicaID > 0:
        for fname, info in entries.items():
            # add replica and register offsets
            offset = info[0]
            data_offset = info[1]
            size = info[2]
            SQLExecute(
                cur,
                "insert into archiveidx (archiveid, replicaid, filename, offset, offset_data, size)"
                " values  (?, ?, ?, ?, ?, ?) "
                "on conflict (archiveid, replicaid, filename) do update set offset = ?, offset_data = ?, size = ?",
                (archiveID, archived_replicaID, fname, offset, data_offset, size, offset, data_offset, size),
            )
        SQLCommit(con)


def ArchiveIdx(args: argparse.Namespace, archiveID: int, cur: sqlite3.Cursor, con: sqlite3.Connection, indent: str = ""):
    try:
        csvfile = open(args.tarfileidx, newline="")
        reader = csv.reader(csvfile)
    except FileNotFoundError:
        raise Exception(f"File '{args.tarfileidx}' not found.")
    except Exception as e:
        raise Exception(f"Error occurred when opening '{args.tarfileidx}': {e}")

    # Find archive dir
    res = SQLExecute(cur, f"select dirid, tarname from archive where rowid = {archiveID}")
    rows = res.fetchall()
    if len(rows) == 0:
        raise Exception(f"Archive ID not found: {archiveID}")

    dirID: int = rows[0][0]
    tarname: str = rows[0][1]
    if not tarname:
        raise Exception(f"Directory.Archive {dirID}.{archiveID} is not a TAR archive.")

    line_number = 0
    readnext = True
    while True:
        if readnext:
            row = next(reader, None)
            if row is not None:
                line_number += 1
        else:
            readnext = True
        if row is None:
            break

        # print(f"{line_number}: {row}")
        if len(row) != 5:
            print(
                f"{indent}  Warning: Line {line_number} in {args.tarfileidx} does not have 5 elements. "
                f"Found {len(row)}. Skip."
            )
            continue
        entrytype = int(row[0].strip())
        if entrytype != 0 and entrytype != 5:  # process only Regular and Directory entries
            continue
        offset = int(row[1].strip())
        data_offset = int(row[2].strip())
        size = int(row[3].strip())
        archivename = row[4].strip()

        # find (first non-deleted) replica of dataset that matches the name
        res = SQLExecute(
            cur,
            f"select rowid, datasetid, hostid, dirid, size from replica where "
            f"name = '{archivename}' and deltime = 0",
        )
        replica_row = res.fetchone()
        if replica_row is None:
            if args.verbose:
                print(f"{indent}  No suitable replica of {archivename} found. Skip")
            continue
        replicaID: int = replica_row[0]
        replica_datasetID: int = replica_row[1]
        replica_hostID: int = replica_row[2]
        replica_dirID: int = replica_row[3]
        replica_size: int = replica_row[4]
        print(f"{indent}Replica id = {replicaID} on host {replica_hostID}, dir {replica_dirID}")

        # find dataset of this replica
        res = SQLExecute(cur, f"select name, fileformat from dataset where rowid = '{replica_datasetID}'")
        dsrow = res.fetchone()
        dsname = dsrow[0]
        format: str = dsrow[1]
        print(f"{indent}  Dataset {replica_datasetID:<5} {dsname}")

        entries: dict = {"": [offset, data_offset, size]}
        if entrytype == TARTYPES["reg"]:
            if size == replica_size:
                ArchiveIdxReplica(dsname, dirID, archiveID, replicaID, entries, cur, con, indent=indent + "  ")
            else:
                print(
                    f"{indent}  The replica size ({replica_size}) does not match the size "
                    f"in the TAR file ({size}). Skip"
                )

        elif entrytype == TARTYPES["dir"] and format == "ADIOS":
            # it's a directory for ADIOS datasets, process its entries
            while True:
                row = next(reader, None)
                if row is None:
                    break
                line_number += 1
                entrytype = int(row[0].strip())
                offset = int(row[1].strip())
                data_offset = int(row[2].strip())
                size = int(row[3].strip())
                entryname: str = row[4].strip()
                if not entryname.startswith(archivename):
                    break
                if entrytype == TARTYPES["reg"]:
                    # a file inside the ADIOS dataset
                    fname = entryname[len(archivename) + 1 :]
                    entries[fname] = [offset, data_offset, size]
            # we have a row unprocessed or None, skip reading at the beginning of the loop
            readnext = False
            ArchiveIdxReplica(dsname, dirID, archiveID, replicaID, entries, cur, con, indent=indent + "  ")
    csvfile.close()


def AddArchivalStorage(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    protocol = args.system.lower()
    if protocol != "https" and protocol != "http" and protocol != "ftp" and protocol != "s3":
        protocol = ""

    print(f"Add archival storage host = {args.host}, directory = {args.directory}, archive system {args.system}")
    print(f"                     tarfile = {args.tarfilename} taridx = {args.tarfileidx}")

    hostID = AddHostName(args.longhostname, args.host, cur, protocol, indent="  ")
    dirID = AddDirectory(hostID, args.directory, cur, indent="  ")
    notes = None
    if args.note:
        try:
            with open(args.note, "rb") as f:
                notes = f.read()
        except IOError as e:
            print(f"WARNING: Failed to read notes from {args.notes}: {e.strerror}.")
            notes = None
    tarname = ""
    if args.tarfilename:
        tarname = args.tarfilename
        print(f"  Adding a TAR file: {tarname}")

    res = SQLExecute(
        cur,
        "select rowid from archive where dirid = " + str(hostID) + ' and tarname = "' + tarname + '"',
    )
    row = res.fetchone()
    if row is not None:
        archiveID = row[0]
        print(f"  Found archive already in the database, rowid = {archiveID}")
    else:
        curArchive = SQLExecute(
            cur,
            "insert into archive (dirid, tarname, system, notes) values  (?, ?, ?, ?) ",
            (dirID, tarname, args.system, notes),
        )
        archiveID = lastrowid_or_zero(curArchive)
        SQLCommit(con)

    if archiveID == 0:
        print("  ERROR: Could not insert information into table 'archive' for some reason")
    elif args.tarfileidx:
        ArchiveIdx(args, archiveID, cur, con, indent="  ")


def Update(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    longHostName, shortHostName = GetHostName(args)

    hostID = AddHostName(longHostName, shortHostName, cur)
    keyID = AddKeyID(args.encryption_key_id, cur)

    if args.remote_data and args.s3_bucket is not None:
        rootdir = args.s3_bucket
    else:
        rootdir = getcwd()

    dirID = AddDirectory(hostID, rootdir, cur)
    SQLCommit(con)

    if args.command == "dataset":
        ProcessDatasets(args, cur, hostID, dirID, keyID, longHostName + rootdir, rootdir)
    elif args.command == "text":
        ProcessTextFiles(args, cur, hostID, dirID, keyID, longHostName + rootdir, rootdir)
    elif args.command == "image":
        ProcessImage(args, cur, hostID, dirID, keyID, longHostName + rootdir, rootdir)

    SQLCommit(con)


def Create(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    print(f"Create new archive {args.CampaignFileName}")
    SQLExecute(cur, "create table info(id TEXT, name TEXT, version TEXT, modtime INT)")
    SQLCommit(con)
    SQLExecute(
        cur,
        "insert into info values (?, ?, ?, ?)",
        ("ACA", "ADIOS Campaign Archive", ACA_VERSION, CURRENT_TIME),
    )

    SQLExecute(cur, "create table key" + "(keyid TEXT PRIMARY KEY)")
    SQLExecute(
        cur,
        "create table host"
        + "(hostname TEXT PRIMARY KEY, longhostname TEXT, modtime INT, deltime INT, default_protocol TEXT)",
    )
    SQLExecute(
        cur,
        "create table directory" + "(hostid INT, name TEXT, modtime INT, deltime INT, PRIMARY KEY (hostid, name))",
    )
    SQLExecute(
        cur,
        "create table timeseries" + "(tsid INTEGER PRIMARY KEY, name TEXT UNIQUE)",
    )
    SQLExecute(
        cur,
        "create table dataset"
        + "(name TEXT, uuid TEXT, modtime INT, deltime INT, fileformat TEXT, tsid INT, tsorder INT"
        + ", PRIMARY KEY (name))",
    )
    SQLExecute(
        cur,
        "create table replica"
        + "(datasetid INT, hostid INT, dirid INT, archiveid INT, name TEXT, modtime INT, deltime INT"
        + ", keyid INT, size INT"
        + ", PRIMARY KEY (datasetid, hostid, dirid, archiveid, name))",
    )
    SQLExecute(
        cur,
        "create table file"
        + "(replicaid INT, name TEXT, compression INT, lenorig INT"
        + ", lencompressed INT, modtime INT, checksum TEXT, data BLOB"
        + ", PRIMARY KEY (replicaid, name))",
    )
    SQLExecute(
        cur,
        "create table accuracy" + "(replicaid INT, accuracy REAL, norm REAL, relative INT, PRIMARY KEY (replicaid))",
    )
    SQLExecute(cur, "create table resolution" + "(replicaid INT, x INT, y INT, PRIMARY KEY (replicaid))")
    SQLExecute(
        cur, "create table archive" + "(dirid INT, tarname TEXT,system TEXT, notes BLOB, PRIMARY KEY (dirid, tarname))"
    )
    SQLExecute(
        cur,
        "create table archiveidx"
        + "(archiveid INT, replicaid INT, filename TEXT, offset INT, offset_data INT, size INT"
        + ", PRIMARY KEY (archiveid, replicaid, filename))",
    )
    SQLCommit(con)
    cur.close()
    con.close()
    while not exists(args.CampaignFileName):
        sleep(0.1)


def DeleteDatasetIfEmpty(
    args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection, datasetid: int, indent: str
):
    print(f"{indent}Check if dataset {datasetid} still has replicas")
    res = SQLExecute(cur, "select rowid from replica " + f" where datasetid = {datasetid} and deltime = 0")
    replicas = res.fetchall()
    if len(replicas) == 0:
        print("{indent}  Dataset without replicas found. Deleting.")
        SQLExecute(cur, f"update dataset set deltime = {CURRENT_TIME} " + f"where rowid = {datasetid}")


def DeleteReplica(
    args: argparse.Namespace,
    cur: sqlite3.Cursor,
    con: sqlite3.Connection,
    repid: int,
    delete_empty_dataset: bool,
    indent: str = "",
):
    print(f"{indent}Delete replica with id {repid}")
    res = SQLExecute(cur, "select datasetid, hostid, dirid from replica " + f"where rowid = {repid}")
    replicas = res.fetchall()
    datasetid = 0
    for rep in replicas:
        datasetid = rep[0]
        SQLExecute(cur, f"update replica set deltime = {CURRENT_TIME} " + f"where rowid = {repid}")
    if delete_empty_dataset:
        SQLExecute(cur, f"delete from file where replicaid = {repid}")
        DeleteDatasetIfEmpty(args, cur, con, datasetid, indent=indent + "  ")


def DeleteDataset(
    args: argparse.Namespace,
    cur: sqlite3.Cursor,
    con: sqlite3.Connection,
    name: str = "",
    uniqueid: str = "",
):
    if len(name) > 0:
        print(f"Delete dataset with name {name}")
        curDS = SQLExecute(
            cur,
            f"update dataset set deltime = {CURRENT_TIME} " f'where name = "{name}" returning rowid',
        )
    elif len(uniqueid) > 0:
        print(f"Delete dataset with uuid = {uniqueid}")
        curDS = SQLExecute(
            cur,
            f"update dataset set deltime = {CURRENT_TIME} " f'where uuid = "{uniqueid}" returning rowid',
        )
    else:
        raise Exception("DeleteDataset() requires name or unique id")

    rowID = curDS.fetchone()[0]
    res = SQLExecute(curDS, "select rowid from replica " + f" where datasetid = {rowID} and deltime = 0")
    replicas = res.fetchall()
    for rep in replicas:
        DeleteReplica(args, cur, con, rep[0], False)


def Delete(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    if args.uuid is not None:
        for uid in args.uuid:
            DeleteDataset(args, cur, con, uniqueid=uid)
            SQLCommit(con)

    if args.name is not None:
        for name in args.name:
            DeleteDataset(args, cur, con, name=name)
            SQLCommit(con)

    if args.replica is not None:
        for repid in args.replica:
            DeleteReplica(args, cur, con, repid, True)
            SQLCommit(con)


def InfoDataset(
    args: argparse.Namespace, dataset: list, cur: sqlite3.Cursor, delete_condition_and: str, dirs_archived: list[bool]
):
    datasetID = dataset[0]
    timestr = timestamp_to_str(dataset[3])
    print(f"    {dataset[1]}  {dataset[5]:5}  {timestr}   {dataset[2]}", end="")
    if dataset[4] > 0:
        print(f"  - deleted {timestamp_to_str(dataset[4])}")
    else:
        print()

    if not args.list_replicas and not args.list_files:
        return

    res2 = SQLExecute(
        cur,
        "select rowid, hostid, dirid, archiveid, name, modtime, deltime, keyid, size from replica "
        + 'where datasetid = "'
        + str(datasetID)
        + '"'
        + delete_condition_and,
    )
    replicas = res2.fetchall()
    for rep in replicas:
        replicaid = rep[0]
        # hostid = rep[1]
        dirid = rep[2]
        archiveid = rep[3]
        name = rep[4]
        modtime = rep[5]
        deltime = rep[6]
        keyid = rep[7]
        size = rep[8]
        if deltime > 0 and not args.show_deleted:
            return

        flagDel = "-"
        flagRemote = "r"
        flagEncrypted = "-"
        flagAccuracy = "-"
        flagArchive = "-"
        if deltime > 0:
            flagDel = "D"

        if keyid > 0:
            flagEncrypted = "k"

        if dirs_archived[dirid]:
            flagArchive = "A"

        if dataset[5] == "ADIOS" or dataset[5] == "HDF5":
            res3 = SQLExecute(cur, f"select rowid from accuracy where replicaid = {replicaid}")
            acc = res3.fetchall()
            if len(acc) > 0:
                flagAccuracy = "a"

        if dataset[5] == "IMAGE" or dataset[5] == "TEXT":
            res3 = SQLExecute(cur, f"select rowid from file where replicaid = {replicaid}")
            res = res3.fetchall()
            if len(res) > 0:
                flagRemote = "e"

        timestr = timestamp_to_str(modtime)
        print(
            f"  {replicaid:>7} {flagRemote}{flagEncrypted}{flagAccuracy}{flagArchive}{flagDel} {dirid}",
            end="",
        )
        if archiveid > 0:
            print(f".{archiveid}", end="")

        if dataset[5] == "IMAGE":
            res3 = SQLExecute(
                cur,
                'select rowid, x, y from resolution where replicaid = "' + str(replicaid) + '"',
            )
            res = res3.fetchall()
            if len(res) > 0:
                print(f" {res[0][1]} x {res[0][2]}".rjust(14), end="")
        else:
            print(" ".rjust(14), end="")

        print(
            f" {sizeof_fmt(size):>11}  {timestr}",
            end="",
        )

        print(f"      {name}", end="")

        if deltime > 0:
            print(f"  - deleted {timestamp_to_str(deltime)}", end="")

        print()
        if not args.list_files:
            continue

        res3 = SQLExecute(
            cur,
            "select name, lenorig, lencompressed, modtime, checksum from file " + f"where replicaid = {replicaid}",
        )
        files = res3.fetchall()
        for file in files:
            if keyid > 0:
                print("".rjust(28), end="")
                print(f"k{keyid:<3}", end="")
            else:
                print("".rjust(32), end="")
            print(f"{sizeof_fmt(file[2]):>11}  {timestamp_to_str(file[3])}", end="")
            if args.show_checksum:
                print(f"         {file[4]}  {file[0]}", end="")
            else:
                print(f"         {file[0]}", end="")
            print()


def Info(args: argparse.Namespace, cur: sqlite3.Cursor):
    res = SQLExecute(cur, "select id, name, version, modtime from info")
    info = res.fetchone()
    t = timestamp_to_str(info[3])
    print(f"{info[1]}, version {info[2]}, created on {t}")
    print()

    #
    # Hosts and directories
    #
    delete_condition_where = " where deltime = 0"
    delete_condition_and = " and deltime = 0"
    if args.show_deleted:
        delete_condition_where = ""
        delete_condition_and = ""
    print("Hosts and directories:")
    res = SQLExecute(cur, "select rowid, hostname, longhostname from host" + delete_condition_where)
    hosts = res.fetchall()
    dirs_archived = [False]  # [0] is never accessed but needed since dir IDs run 1...n
    for host in hosts:
        # hostID = host[0]
        print(f"  {host[1]}   longhostname = {host[2]}")
        res2 = SQLExecute(
            cur,
            "select rowid, name, modtime, deltime from directory "
            + 'where hostid = "'
            + str(host[0])
            + '"'
            + delete_condition_and,
        )
        dirs = res2.fetchall()
        for dir in dirs:
            if dir[3] == 0 or args.show_deleted:
                # check if it's archive dir
                archive_system = "  "
                res3 = SQLExecute(
                    cur, f"select rowid, tarname, system from archive where dirid = {dir[0]} order by rowid"
                )
                archs = res3.fetchall()
                if len(archs) > 0:
                    archive_system = f"  - Archive: {archs[0][2]}"
                    dirs_archived.append(True)
                else:
                    dirs_archived.append(False)
                print(f"     {dir[0]}. {dir[1]}{archive_system}")
                for idx in range(len(archs)):
                    if archs[idx][1]:
                        print(f"       {dir[0]}.{archs[idx][0]} {archs[idx][1]}")
                    else:
                        print(f"       {dir[0]}.{archs[idx][0]} .")
    print()

    #
    # Keys
    #
    res = SQLExecute(cur, "select rowid, keyid from key")
    keys = res.fetchall()
    if len(keys) > 0:
        print("Encryption keys:")
    for key in keys:
        print(f"  k{key[0]}. {key[1]}")
    if len(keys) > 0:
        print()

    #
    # Time Series
    #
    res = SQLExecute(cur, "select tsid, name from timeseries")
    timeseries = res.fetchall()
    if len(timeseries) > 0:
        print("Time-series and their datasets:")
    for ts in timeseries:
        print(f"  {ts[1]}")
        res = SQLExecute(
            cur,
            "select rowid, uuid, name, modtime, deltime, fileformat from dataset "
            f"where tsid = {ts[0]} " + delete_condition_and,
        )
        datasets = res.fetchall()
        for dataset in datasets:
            InfoDataset(args, dataset, cur, delete_condition_and, dirs_archived)
    if len(timeseries) > 0:
        print("")

    #
    # Datasets
    #
    res = SQLExecute(
        cur,
        "select rowid, uuid, name, modtime, deltime, fileformat from dataset " "where tsid = 0 " + delete_condition_and,
    )
    datasets = res.fetchall()
    if len(datasets) > 0:
        print("Other Datasets:")
    for dataset in datasets:
        InfoDataset(args, dataset, cur, delete_condition_and, dirs_archived)


def DeleteCampaignFile(args: argparse.Namespace):
    if exists(args.CampaignFileName):
        print(f"Delete campaign archive {args.CampaignFileName}")
        remove(args.CampaignFileName)
        while exists(args.CampaignFileName):
            sleep(0.1)
        return 0
    else:
        print(f"ERROR: archive {args.CampaignFileName} does not exist")
        return 1


def CampaignInfo(filename):
    import io
    import sys
    import re

    output = io.StringIO()
    sys.stdout = output
    main(
        args=[filename, "info"],
        prog=None,
    )
    output_string = output.getvalue()
    sys.stdout = sys.__stdout__
    pattern = re.compile(
        r"^\s*(?P<uuid>[0-9a-f]{32})\s+(?P<type>ADIOS|HDF5|TEXT|IMAGE)\s+.*?\s+.*?\s+.*?\s+(?P<path>.*)", re.MULTILINE
    )
    return [m.groupdict() for m in pattern.finditer(output_string)]


def main(args=None, prog=None):
    parser = ArgParser(args=args, prog=prog)
    CheckCampaignStore(parser.args)

    if parser.args.keyfile:
        key = read_key(parser.args.keyfile)
        # ask for password at this point
        parser.args.encryption_key = key.get_decrypted_key()
        parser.args.encryption_key_id = key.id
    else:
        parser.args.encryption_key = None
        parser.args.encryption_key_id = None

    con: sqlite3.Connection
    cur: sqlite3.Cursor
    connected = False

    while parser.parse_next_command():
        print("=" * 70)
        # print(parser.args)
        # print("--------------------------")
        if parser.args.command == "delete" and parser.args.campaign is True:
            DeleteCampaignFile(parser.args)
            continue

        if parser.args.command == "create":
            # print("Create archive")
            if exists(parser.args.CampaignFileName):
                print(f"ERROR: archive {parser.args.CampaignFileName} already exist")
                exit(1)
        else:
            # print(f"{parser.args.command} archive")
            if not exists(parser.args.CampaignFileName):
                print(f"ERROR: archive {parser.args.CampaignFileName} does not exist")
                exit(1)

        if not connected:
            con = sqlite3.connect(parser.args.CampaignFileName)
            cur = con.cursor()
            connected = True

        if parser.args.command == "info":
            Info(parser.args, cur)
            continue
        elif parser.args.command == "create":
            Create(parser.args, cur, con)
            connected = False
            continue
        elif parser.args.command == "dataset" or parser.args.command == "text" or parser.args.command == "image":
            Update(parser.args, cur, con)
            continue
        elif parser.args.command == "delete":
            Delete(parser.args, cur, con)
            continue
        elif parser.args.command == "add-archival-storage":
            AddArchivalStorage(parser.args, cur, con)
        elif parser.args.command == "archived":
            ArchiveDataset(parser.args, cur, con)
        elif parser.args.command == "time-series":
            AddTimeSeries(parser.args, cur, con)
        elif parser.args.command == "upgrade":
            UpgradeACA(parser.args, cur, con)
        else:
            print("This should not happen. " f"Unknown command accepted by argparser: {parser.args.command}")

    if connected:
        cur.close()
        con.close()

    if len(SQLErrorList) > 0:
        print()
        print("!!!! SQL Errors encountered")
        for e in SQLErrorList:
            print(f"  {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
        print("!!!!")
        print()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import glob
import re
import fnmatch
from os.path import exists, isdir, dirname, basename, expanduser

from .config import Config


def List(*patterns, wildcard: bool = False, campaign_store=None):
    args = argparse.Namespace()
    if wildcard:
        args.wildcard = True
    else:
        args.wildcard = False
    args.pattern = []
    for p in patterns:
        args.pattern.append(p)
    args.verbose = 0
    args.campaign_store = None
    args = _SetDefaults(args)
    _CheckCampaignStore(args)
    return _List(args, collect=True, campaign_store=campaign_store)


def _SetupArgs(args=None, prog=None):
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument("pattern", help="filter pattern(s) as regular expressions", default=None, nargs="*")
    parser.add_argument(
        "-w",
        "--wildcard",
        help="Use patterns as path wildcard patterns",
        action="store_true",
        default=False,
    )
    parser.add_argument("-s", "--campaign_store", help="Path to local campaign store", default=None)
    parser.add_argument("-v", "--verbose", help="More verbosity", action="count", default=0)
    args = parser.parse_args(args=args)
    return _SetDefaults(args)


def _SetDefaults(args: argparse.Namespace):
    # default values
    args.user_options = Config()
    args.host_options = args.user_options.read_host_config()

    if args.verbose == 0:
        args.verbose = args.user_options.verbose

    if args.campaign_store is None:
        args.campaign_store = args.user_options.campaign_store_path

    if args.campaign_store is not None:
        while args.campaign_store[-1] == "/":
            args.campaign_store = args.campaign_store[:-1]

    if args.verbose > 0:
        print(f"# Verbosity = {args.verbose}")
        print(f"# Campaign Store = {args.campaign_store}")
        print(f"# pattern(s) = {args.pattern}")
    return args


def _CheckCampaignStore(args):
    if args.campaign_store is not None and not isdir(args.campaign_store):
        print("ERROR: Campaign directory " + args.campaign_store + " does not exist", flush=True)
        exit(1)


def _List(args: argparse.Namespace, collect: bool = True, campaign_store=None) -> list[str]:
    result = []
    path = campaign_store
    if path is None:
        path = args.campaign_store
    if path is None:
        print("ERROR: Set --campaign_store for this command")
        return result

    # List the local campaign store
    acaList = glob.glob(path + "/**/*.aca", recursive=True)
    if len(acaList) == 0:
        print("There are no campaign archives in  " + path)
        return result
    else:
        startCharPos = len(path) + 1
        for f in acaList:
            name = f[startCharPos:]
            matches = False
            if len(args.pattern) == 0:
                matches = True
            else:
                for p in args.pattern:
                    if args.wildcard:
                        if fnmatch.fnmatch(name, p):
                            matches = True
                            break
                    else:
                        if re.search(p, name):
                            matches = True
                            break

            if matches:
                if collect:
                    result.append(f[startCharPos:])
                else:
                    print(f[startCharPos:])
    return result


def main(args=None, prog=None):
    args = _SetupArgs(args=args, prog=prog)
    _CheckCampaignStore(args)
    _List(args, collect=False)


if __name__ == "__main__":
    main()



import argparse
import sqlite3
from datetime import datetime, timedelta
from .config import ACA_VERSION
from .utils import SQLExecute, SQLCommit, SQLErrorList


def _upgrade_to_0_6(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    print("Upgrade to 0.6")
    # host
    SQLExecute(cur, "ALTER TABLE host ADD default_protocol TEXT")
    # replica
    SQLExecute(
        cur,
        "CREATE TABLE replica_new"
        + "(datasetid INT, hostid INT, dirid INT, archiveid INT, name TEXT, modtime INT, deltime INT"
        + ", keyid INT, size INT"
        + ", PRIMARY KEY (datasetid, hostid, dirid, archiveid, name))",
    )
    SQLExecute(cur, 
        'INSERT INTO replica_new (datasetid, hostid, dirid, name, modtime, deltime, keyid, size)'
        ' SELECT datasetid, hostid, dirid, name, modtime, deltime, keyid, size FROM replica'
    )
    SQLExecute(cur, "UPDATE replica_new SET archiveid = 0")
    SQLExecute(cur, "DROP TABLE replica")
    SQLExecute(cur, "ALTER TABLE replica_new RENAME TO replica")
    # archive
    SQLExecute(
        cur,
        "CREATE TABLE archive_new"
        + "(dirid INT, tarname TEXT, system TEXT, notes BLOB, PRIMARY KEY (dirid, tarname))"
    )
    SQLExecute(cur, 
        'INSERT INTO archive_new (dirid, system, notes)'
        ' SELECT dirid, system, notes FROM archive'
    )
    SQLExecute(cur, 'UPDATE archive_new SET tarname = ""')
    SQLExecute(cur, "DROP TABLE archive")
    SQLExecute(cur, "ALTER TABLE archive_new RENAME TO archive")
    # archiveidx
    SQLExecute(
        cur,
        "create table archiveidx"
        + "(archiveid INT, replicaid INT, filename TEXT, offset INT, offset_data INT, size INT"
        + ", PRIMARY KEY (archiveid, replicaid, filename))",
    )
    # info: update version
    SQLExecute(cur, 'UPDATE info SET version = "0.6"')
    if len(SQLErrorList) == 0:
        SQLCommit(con)
    else:
        print("SQL Errors detected, drop all changes.")


UPGRADESTEP = {
    "0.5": {"new_version": "0.6", "func": _upgrade_to_0_6}
}


def UpgradeACA(args: argparse.Namespace, cur: sqlite3.Cursor, con: sqlite3.Connection):
    res = SQLExecute(cur, 'select version from info where id = "ACA"')
    info = res.fetchone()
    version: str = info[0]
    if version != ACA_VERSION:
        print(f"Current version is {version}")
        # vlist = version.split('.')
        v = UPGRADESTEP.get(version)
        if v is not None:
            v["func"](args, cur, con)
        else:
            print("This version cannot be upgraded")
    else:
        print(f"This archive has the latest version already: {ACA_VERSION}")


 


#!/usr/bin/env python3
import argparse
import sqlite3
import redis
from os import walk, listdir
from shutil import rmtree
from re import match
from glob import glob
from os.path import exists, join, getsize
from sys import exit
from time import sleep

import redis.exceptions

from .config import Config, REDIS_PORT
from .utils import timestamp_to_datetime, input_yes_or_no


def setup_args(cfg: Config, args=None, prog=None):
    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument(
        "command",
        help="Command: list/clear",
        choices=["list", "clear"],
    )
    parser.add_argument("campaign", help="Campaign name or path, with .aca or without", default=None, nargs="?")
    parser.add_argument("--redis-port", "-p", help="Key-value database port", default=REDIS_PORT)
    parser.add_argument("--verbose", "-v", help="More verbosity", action="count", default=0)
    parser.add_argument("--yes-to-all", "-y", help="Answer yes automatically", action="store_true", default=False)
    args = parser.parse_args(args=args)

    args.CampaignFileName = args.campaign
    if args.campaign is not None:
        if not args.campaign.endswith(".aca"):
            args.CampaignFileName += ".aca"
        if (
            not exists(args.CampaignFileName)
            and not args.CampaignFileName.startswith("/")
            and cfg.campaign_store_path is not None
        ):
            args.CampaignFileName = cfg.campaign_store_path + "/" + args.CampaignFileName

    if args.verbose > 1:
        print(f"# Verbosity = {args.verbose}")
        print(f"# Command = {args.command}")
        print(f"# REDIS port = {args.redis_port}")
        print(f"# Archive = {args.CampaignFileName}")
        print(f"# Auto yes = {args.yes_to_all}")
    return args


def folder_size(folder_path: str) -> int:
    folder_size = 0
    for path, dirs, files in walk(folder_path):
        for f in files:
            fp = join(path, f)
            folder_size += getsize(fp)
    return folder_size


def list_cache(args: argparse.Namespace, cfg: Config, kvdb: redis.Redis):
    archives = {}  # organize datasets to archives
    cache_folders = glob("[0-9a-f][0-9a-f][0-9a-f]", root_dir=cfg.cache_path)
    if args.verbose > 1:
        print(f"# Found {len(cache_folders)} cache folders in cache directory")
    for folder in cache_folders:
        folder_path = join(cfg.cache_path, folder)
        dataset_ids = glob("[0-9a-f]*", root_dir=folder_path)
        if args.verbose > 1:
            print(f"# Found {len(dataset_ids)} datasets in cache folder {folder}")
        for id in dataset_ids:
            archive_name = "unknown"
            infoname = join(folder_path, id, "info.txt")
            if exists(infoname):
                infofile = open(infoname, "r")
                for line in infofile:
                    if match("Campaign = ", line):
                        archive_name = line[11:-1]
            dirsize = folder_size(join(folder_path, id))

            kvkeys = kvdb.keys(id + "*")
            nkv = len(kvkeys)
            kvsize = 0
            for key in kvkeys:
                kvsize += kvdb.memory_usage(key)

            if args.verbose > 1:
                print(f"# {id} from archive {archive_name}, cache size = {dirsize}, # of keys = {nkv}")
            entry = {id: {"dirsize": dirsize, "nkv": nkv, "kvsize": kvsize}}
            if archive_name not in archives:
                archives[archive_name] = {}
            archives[archive_name].update(entry)
    if args.verbose > 1:
        print("")

    print("folder-size     db-entries db-size     campaign name")
    print("--------------------------------------------------------------------------")
    size_all = 0
    nkv_all = 0
    kvsize_all = 0
    for arch in archives:
        size_arch = 0
        nkv_arch = 0
        kvsize_arch = 0
        for id, idvalues in archives[arch].items():
            size_arch += idvalues["dirsize"]
            nkv_arch += idvalues["nkv"]
            kvsize_arch += idvalues["kvsize"]
        print(f"{size_arch:<15} {nkv_arch:<10} {kvsize_arch:<11} {arch}")
        size_all += size_arch
        nkv_all += nkv_arch
        kvsize_all += kvsize_arch
        if args.verbose > 0:
            for id, idvalues in archives[arch].items():
                print(f"{idvalues['dirsize']:>14}   {idvalues['nkv']:>8}  " f"{idvalues['kvsize']:>10}     {id}")
    print(f"{size_all:<15} {nkv_all:<10} {kvsize_all}")


def delete_cache_items(args: argparse.Namespace, cfg: Config, kvdb: redis.Redis, id: str):
    kvkeys = kvdb.keys(id + "*")
    nkeys = len(kvkeys)
    parent_path = join(cfg.cache_path, id[0:3])
    path = join(parent_path, id)

    if nkeys > 0 or exists(path):
        if args.yes_to_all or input_yes_or_no("Do you want to clear cache for " + id + " (y/n)? "):
            # delete KV entries
            if nkeys > 0:
                kvdb.delete(*kvkeys)
            if args.verbose > 0:
                print(f"  deleted {nkeys} keys from cache db")

            # delete files
            rmtree(path, ignore_errors=True)
            if args.verbose > 0:
                print(f"  deleted folder {path}")
    else:
        if args.verbose > 0:
            print("  nothing in cache")

    # delete cache_path/XXX is empty
    if exists(parent_path):
        dir = listdir(parent_path)
        if len(dir) == 0:
            rmtree(parent_path, ignore_errors=True)
            if args.verbose > 0:
                print(f"  deleted folder {parent_path}")


def clear_cache(args: argparse.Namespace, cfg: Config, kvdb: redis.Redis):
    con = sqlite3.connect(args.CampaignFileName)
    cur = con.cursor()

    res = cur.execute("select id, name, version, modtime from info")
    info = res.fetchone()
    t = timestamp_to_datetime(info[3])
    print(f"{info[1]}, version {info[2]}, created on {t}")

    res = cur.execute("select rowid, uuid, name, modtime from dataset")
    datasets = res.fetchall()
    for dataset in datasets:
        id = dataset[1]
        t = timestamp_to_datetime(dataset[3])
        print(f"        dataset = {id}    {t}    {dataset[2]} ")
        delete_cache_items(args, cfg, kvdb, id)

    cur.close()
    con.close()


def connect_to_redis(host: str, port: int, db: int) -> redis.Redis | None:
    r = redis.Redis(host=host, port=port, db=db)
    try:
        r.ping()
    except (redis.exceptions.ConnectionError, ConnectionRefusedError):
        print(f"Could not connect to Redis at {host}:{port}, db={db}. Check if Redis is running.")
        return None
    return r


def main(args=None, prog=None):
    # default values
    cfg = Config()
    args = setup_args(cfg, args=args, prog=prog)
    if not cfg.cache_path:
        print("No cachepath specified in user config")
        exit(1)

    if not exists(cfg.cache_path):
        print(f"Could not find {cfg.cache_path}")
        exit(1)

    kvdb = connect_to_redis(host="localhost", port=args.redis_port, db=0)
    if not kvdb:
        exit(1)

    if args.command == "list":
        if args.CampaignFileName is not None:
            print("Ignoring campaign archive argument")
        list_cache(args, cfg, kvdb)

    elif args.command == "clear":
        if args.CampaignFileName is None:
            print("Missing campaign archive argument for clearing cache")
            exit(1)
        clear_cache(args, cfg, kvdb)


if __name__ == "__main__":
    main()



import nacl.encoding
import nacl.secret
import nacl.utils
import nacl.pwhash
import uuid
import yaml
from datetime import datetime, UTC
from getpass import getpass
from os.path import exists


class Key:
    """Key file for encryption of campaign metadata"""

    def __init__(self):
        self.id = None
        self.password_encrypted = False
        self.key = None
        self.date = datetime.now(UTC)
        self.note = ""
        self.salt = None

    def generate(self, note: str, pwd: bytes = None):
        self.date = datetime.now(UTC)
        self.note = note

        # Generate a secret key
        unencrypted_key = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)

        # Generate a UUID for the key
        self.id = uuid.uuid4().hex

        if pwd:
            kdf = nacl.pwhash.argon2i.kdf
            self.salt = nacl.utils.random(nacl.pwhash.argon2i.SALTBYTES)
            pkey = kdf(
                nacl.secret.SecretBox.KEY_SIZE,
                pwd,
                self.salt,
                opslimit=nacl.pwhash.argon2i.OPSLIMIT_SENSITIVE,
                memlimit=nacl.pwhash.argon2i.MEMLIMIT_SENSITIVE,
            )
            pbox = nacl.secret.SecretBox(pkey)
            pnonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)
            self.key = pbox.encrypt(unencrypted_key, pnonce)
        else:
            self.key = unencrypted_key
            self.salt = None

    def generate_interactive(self, password_required: bool = False):
        print("Type a note for this key: ", end="")
        note = input()
        if password_required:
            while True:
                pwd = bytes(getpass("Create password: "), "utf-8")
                pwdcheck = bytes(getpass("Retype password: "), "utf-8")
                if pwd != pwdcheck:
                    print("Password mismatch. Try again.")
                else:
                    break
        else:
            pwd = None
        self.generate(note, pwd)

    def write(self, path: str):
        doc = {}
        doc["note"] = self.note
        doc["id"] = self.id
        doc["key"] = self.key.hex()
        if self.salt:
            doc["salt"] = self.salt.hex()
        doc["date"] = self.date.isoformat()
        with open(path, "w") as file:
            yaml.dump(doc, file, sort_keys=False)

    def read(self, path: str):
        with open(path, "r") as file:
            doc = yaml.safe_load(file)

        self.note = doc["note"]
        self.id = doc["id"]
        self.password_encrypted = False
        self.key = bytes.fromhex(doc["key"])
        self.date = doc["date"]

        if "salt" in doc:
            self.salt = bytes.fromhex(doc["salt"])
        else:
            self.salt = None

    def get_decrypted_key(self, pwd: bytes = None) -> bytes:
        if self.salt:
            if not pwd:
                pwd = bytes(getpass("Password: "), "utf-8")
            kdf = nacl.pwhash.argon2i.kdf
            pkey = kdf(
                nacl.secret.SecretBox.KEY_SIZE,
                pwd,
                self.salt,
                opslimit=nacl.pwhash.argon2i.OPSLIMIT_SENSITIVE,
                memlimit=nacl.pwhash.argon2i.MEMLIMIT_SENSITIVE,
            )
            pbox = nacl.secret.SecretBox(pkey)
            return pbox.decrypt(self.key)
        else:
            return self.key

    def info(self, do_verify: bool = False) -> bool:
        print(f"created on: {datetime.fromisoformat(self.date)}")
        print(f"      note: {self.note}")
        print(f"      uuid: {self.id}")
        if self.salt:
            print("      encryption: password")
        else:
            print("      encryption: none")

        if do_verify and self.salt:
            try:
                _ = self.get_decrypted_key()
            except Exception:
                print("Password is incorrect")
                return False
            print("Password is correct")
        return True


def read_key(path: str) -> Key:
    if not exists(path):
        print(f"Could not find key file {path}")
        exit(1)
    key = Key()
    key.read(path)
    return key



try:
    import h5py

    __HAVE_H5PY__ = True
except Exception:
    __HAVE_H5PY__ = False

import numpy as np
import sys
from os import stat
from os.path import exists


def _report(operation, key, obj, size):
    type_str = type(obj).__name__.split(".")[-1].lower()
    print(f"{operation} {type_str} {key}, size = {size}")


def h5py_compatible_attributes(in_object):
    """Are all attributes of an object readable in h5py?"""
    try:
        # Force obtaining the attributes so that error may appear
        [0 for at in in_object.attrs.items()]
        return True
    except Exception:
        return False


def copy_attributes(in_object, out_object):
    """Copy attributes between 2 HDF5 objects."""
    for key, value in in_object.attrs.items():
        out_object.attrs[key] = value


def walk(in_object, out_object, log=False):
    """Recursively copy the tree but don't write actual array content

    If attributes cannot be transferred, a copy is created.
    Otherwise, dataset are compressed.
    """
    keys = list(in_object.keys())
    for key in keys:
        in_obj = in_object[key]
        if not isinstance(in_obj, h5py.Datatype) and h5py_compatible_attributes(in_obj):
            if isinstance(in_obj, h5py.Group):
                out_obj = out_object.create_group(key)
                walk(in_obj, out_obj, log)
                if log:
                    _report("Copied", key, in_obj, 0)
            elif isinstance(in_obj, h5py.Dataset):
                if in_obj.size <= 128:
                    out_obj = out_object.create_dataset(key, data=in_obj)
                    if log:
                        _report("Copied small object", key, in_obj, in_obj.size)
                else:
                    out_obj = out_object.create_dataset_like(key, other=in_obj)
                    if log:
                        _report("Recreated without copy", key, in_obj, in_obj.size)
            else:
                raise Exception("Invalid object type %s" % type(in_obj))
            copy_attributes(in_obj, out_obj)
        else:
            # We copy datatypes and objects with non-understandable attributes
            # identically.
            if log:
                _report("Copied object with non-understandable attributes", key, in_obj, 0)
            in_object.copy(key, out_object)


def copy_hdf5_file_without_data(infilename: str, outfilename: str, log: bool = False):
    """Copy an HDF5 file metadata without writing the array data

    :param infilename: Input HDF5 path
    :param outfilename: Output HDF5 path
    :param log: Whether to print results of operations'
    :returns: A tuple(original_size, new_size)
    """
    if __HAVE_H5PY__:
        with h5py.File(infilename, "r") as in_file, h5py.File(outfilename, "w") as out_file:
            walk(in_file, out_file, log=log)
        return stat(infilename).st_size, stat(outfilename).st_size
    else:
        return 0, 0


def IsHDF5Dataset(dataset):
    it_is = False
    if __HAVE_H5PY__:
        try:
            with h5py.File(dataset, "r") as _:
                it_is = True
        except Exception:
            it_is = False
    return it_is


def main(args=sys.argv[1:3], prog=None):
    infilename, outfilename = args
    insize, outsize = copy_hdf5_file_without_data(infilename, outfilename, log=True)
    print(f"{infilename} size = {insize}")
    print(f"{outfilename} size = {outsize}")


if __name__ == "__main__":
    main()



from datetime import datetime, timedelta
from time import sleep
from os import walk
from os.path import join, getsize
import sqlite3

def timestamp_to_datetime(timestamp: int) -> datetime:
    digits = len(str(int(timestamp)))
    t = float(timestamp)
    if digits > 18:
        t = t / 1000000000
    elif digits > 15:
        t = t / 1000000
    elif digits > 12:
        t = t / 1000
    return datetime.fromtimestamp(t)


def datetime_to_str(t: datetime) -> str:
    # The constant 31556952 is used in the ls source code,
    # available at https://www.gnu.org/software/coreutils/.
    # It roughly represents the number of seconds in a Gregorian year.
    six_months_in_seconds = 31556952 // 2
    if (datetime.now() - t) < timedelta(seconds=six_months_in_seconds):
        date_format = "%b %e %H:%M"
    else:
        date_format = "%b %e  %Y"
    return t.strftime(date_format)
    # return t.strftime('%Y-%m-%d %H:%M:%S')


def timestamp_to_str(timestamp: int) -> str:
    t = timestamp_to_datetime(timestamp)
    return datetime_to_str(t)


def input_yes_or_no(msg: str, default_answer: bool = False) -> bool:
    ret = default_answer
    print(msg, end="")
    while True:
        answer = input().lower()
        if answer == "n" or answer == "no":
            ret = False
            break
        if answer == "y" or answer == "yes":
            ret = True
            break
        print("Answer y[es] or n[o]: ", end="")
    return ret


def get_folder_size(Folderpath: str) -> int:
    size = 0
    for path, dirs, files in walk(Folderpath):
        for f in files:
            size += getsize(join(path, f))
    return size


def sizeof_fmt(num: int, suffix="B") -> str:
    if num < 1024:
        return f"{num}"
    n = float(num) / 1024
    for unit in ("Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(n) < 1024.0:
            return f"{n:3.1f} {unit}{suffix}"
        n /= 1024.0
    return f"{n:.1f} Yi{suffix}"


SQLErrorList = []


def SQLExecute(cur: sqlite3.Cursor, cmd: str, parameters=()) -> sqlite3.Cursor:
    res = cur
    try:
        res = cur.execute(cmd, parameters)
    except sqlite3.OperationalError as e:
        print(f"SQL execute Operational Error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
        SQLErrorList.append(e)
        sleep(1.0)
        try:
            res = cur.execute(cmd, parameters)
        except sqlite3.Error as e:
            print(f"SQL re-execute error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
            raise e
        else:
            print("SQL re-execute succeeded")

    except sqlite3.Error as e:
        print(f"SQL execute Error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
        raise e
    finally:
        return res


def SQLCommit(con: sqlite3.Connection):
    try:
        con.commit()
    except sqlite3.OperationalError as e:
        print(f"SQL commit Operational Error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
        SQLErrorList.append(e)
        if e.sqlite_errorcode == sqlite3.SQLITE_IOERR_DELETE:
            sleep(1.0)
            try:
                con.commit()
            except sqlite3.Error as e:
                print(f"SQL recommit error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
                raise e
            else:
                print("SQL recommit succeeded")
    except sqlite3.Error as e:
        print(f"SQL commit Error: {e.sqlite_errorcode}  {e.sqlite_errorname}: {e}")
        raise e



import argparse
import tarfile

TARTYPES = {
    "reg": 0,
    "lnk": 1,
    "sym": 2,
    "chr": 3,
    "blk": 4,
    "dir": 5,
    "fifo": 6,
    "cont": 7,
    "longname": 8,
    "longlink": 9,
    "sparse": 10,
}

def CreateTarIndex(tarfilename: str, indexfile: str | None):
    tf = tarfile.open(tarfilename)
    if indexfile is None:
        indexfile = tarfilename + ".idx"
    with open(indexfile, "w") as idxf:
        for ti in tf:
            idxf.write(f'{int(ti.type)},{ti.offset},{ti.offset_data},{ti.size},"{ti.name}"\n')

def _SetupArgs(args=None, prog=None):
    parser = argparse.ArgumentParser(
        prog=prog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
Create an index file from a TAR file. It can be used in the 
'manager <archive> add-archival-storage' command to make automatic 
replicas of all datasets in the archive and register their offsets 
and sizes in the TAR file.
""",
        )
    parser.add_argument("tarfile", help="Name of the TAR file", type=str)
    parser.add_argument("idxfile", nargs="?", help="Optional name of the index file", type=str)
    args = parser.parse_args(args=args)
#    if args.idxfile is None:
#        args.idxfile = args.tarfile + ".idx"
    return args


def main(args=None, prog=None):
    args = _SetupArgs(args=args, prog=prog)
    CreateTarIndex(args.tarfile, args.idxfile)




#!/usr/bin/env python3

import argparse
import sys
from os.path import exists, basename

from .config import Config

__accepted_commands__ = [
    "create",
    "delete",
    "info",
    "dataset",
    "text",
    "image",
    "add-archival-storage",
    "archived",
    "time-series",
    "upgrade"
]
__accepted_commands_str__ = " | ".join(__accepted_commands__)

prog = basename(sys.argv[0])


class ArgParser:
    """
    Process command-line arguments for the campaign manager

    Usage:
        from hpc_campaign_manager_args import Args
        parser = ArgParser()
        while parser.parse_next_command():
            ... use x.args for argparse namespace

    """

    def __init__(self, args=sys.argv[1:], prog=prog):
        self.parsers = self.setup_args(prog=prog)
        self.commandlines = self.divide_cmdline(__accepted_commands__, args=args)
        self.args = self.parse_args_main(self.parsers["main"], self.commandlines[0])
        self.cmdidx = 1
        self.prev_command = None

    def parse_next_command(self) -> bool:
        if self.cmdidx < len(self.commandlines):
            if self.prev_command is not None:
                self.remove_prev_args(self.prev_command, self.args)
            cmdline = self.commandlines[self.cmdidx]
            self.args = self.parse_args_command(self.args, self.parsers[cmdline[0]], cmdline)
            self.cmdidx += 1
            if self.args.command == "dataset":
                self.check_args_dataset(self.args)
            elif self.args.command == "text":
                self.check_args_text(self.args)
            self.prev_command = cmdline[0]
            return True
        else:
            return False

    def divide_cmdline(self, commands: list, args=sys.argv[1:]):
        # Divide argv by commands
        split_argv = [[]]
        for c in args:
            if c in commands:
                split_argv.append([c])
            else:
                split_argv[-1].append(c)
        return split_argv

    def parse_args_main(self, parser, argv) -> argparse.Namespace:
        args = parser.parse_args(argv)  # Without command

        # default values
        args.user_options = Config()
        args.host_options = args.user_options.read_host_config()

        if args.verbose == 0:
            args.verbose = args.user_options.verbose

        if args.campaign_store is None:
            args.campaign_store = args.user_options.campaign_store_path

        if args.campaign_store is not None:
            while args.campaign_store[-1] == "/":
                args.campaign_store = args.campaign_store[:-1]

        args.remote_data = False
        args.s3_endpoint = None
        if args.hostname is None:
            args.hostname = args.user_options.host_name
        elif args.hostname in args.host_options and args.hostname != args.user_options.host_name:
            args.remote_data = True
            hostopt = args.host_options.get(args.hostname)
            if hostopt is not None:
                optID = next(iter(hostopt))
                if hostopt[optID]["protocol"].casefold() == "s3":
                    args.s3_endpoint = hostopt[optID]["endpoint"]
                    if args.s3_bucket is None:
                        print("ERROR: Remote option for an S3 server requires --s3_bucket")
                        exit(1)
                    if args.s3_datetime is None:
                        print("ERROR: Remote option for an S3 server requires --s3_datetime")
                        exit(1)

        args.CampaignFileName = args.archive
        if args.archive is not None:
            if not args.archive.endswith(".aca"):
                args.CampaignFileName += ".aca"
            if (
                not exists(args.CampaignFileName)
                and not args.CampaignFileName.startswith("/")
                and args.campaign_store is not None
            ):
                args.CampaignFileName = args.campaign_store + "/" + args.CampaignFileName

        args.LocalCampaignDir = ".adios-campaign/"

        if args.verbose > 0:
            print(f"# Verbosity = {args.verbose}")
            print(f"# Campaign File Name = {args.CampaignFileName}")
            print(f"# Campaign Store = {args.campaign_store}")
            print(f"# Host name = {args.hostname}")
            print(f"# Key file = {args.keyfile}")

        return args

    def parse_args_command(self, args: argparse.Namespace, parser, argv) -> argparse.Namespace:
        # Parse one command
        # n = argparse.Namespace()
        # setattr(args, argv[0], n)
        args.command = argv[0]
        parser.parse_args(argv[1:], namespace=args)
        return args

    def check_args_dataset(self, args):
        if args.name is not None:
            if len(args.files) > 1:
                raise Exception(
                    "Invalid arguments for dataset: when using --name <name>, " "only one dataset is allowed"
                )

    def check_args_text(self, args):
        if args.name is not None:
            if len(args.files) > 1:
                raise Exception(
                    "Invalid arguments for text: when using --name <name>, " "only one text file is allowed"
                )

    def setup_args(self, prog=prog) -> dict:
        parser = argparse.ArgumentParser(
            prog=prog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog="""
A campaign archive name without '.aca' extension will be forced to have '.aca'.
If it exists, 'campaignstorepath' in ~/.config/hpc-campaign/config.yaml will be used for
relative paths for <archive> names.
Multiple commands can be used in one run.
Type '%(prog)s <archive> <command> -h' for help on commands.
""",
        )
        parsers = {}
        parsers["main"] = parser
        parser.add_argument("--verbose", "-v", help="More verbosity", action="count", default=0)
        parser.add_argument("--campaign_store", "-s", help="Path to local campaign store", default=None)
        parser.add_argument("--hostname", "-n", help="Host name unique for hosts in a campaign")
        parser.add_argument("--keyfile", "-k", help="Key file to encrypt metadata")
        # parser.add_argument("--s3_bucket", help="Bucket on S3 server", default=None)
        # parser.add_argument(
        #     "--s3_datetime",
        #     help="Datetime of data on S3 server in " "'2024-04-19 10:20:15 -0400' format",
        #     default=None,
        # )
        parser.add_argument("archive", help="Campaign archive name or path, with .aca or without", default=None)
        parser.add_argument("command", nargs="?", help=__accepted_commands_str__, default=None)

        # parser for the "create" command
        parser_create = argparse.ArgumentParser(
            prog=f"{prog} <archive> create",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""Create a new campaign archive file.""",
        )
        parsers["create"] = parser_create

        # parser for the "delete" command
        parser_delete = argparse.ArgumentParser(
            prog=f"{prog} <archive> delete",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""Delete items from a campaign archive file.""",
        )
        parser_delete.add_argument("--uuid", nargs="+", metavar="id", help="Remove datasets by UUID")
        parser_delete.add_argument("--name", nargs="+", metavar="str", help="Remove datasets by name")
        parser_delete.add_argument("--replica", nargs="+", metavar="id", help="Remove replicas by ID number")
        parser_delete.add_argument("--campaign", help="Delete entire campaign file", action="store_true")

        parsers["delete"] = parser_delete

        # parser for the "info" command
        parser_info = argparse.ArgumentParser(
            prog=f"{prog} <archive> info",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""Print content of a campaign archive file.""",
            epilog="""
Listing format:
dataset:  uuid  type  date  representation-name
replica:  replica-id  modes  directory-id  [resolution]  size  date  name
file:     [encryption key]  size  date  [cheksum]  filename
  modes:  e - embedded in archive
          r - remote dataset
          k - encrypted with key (only inside the archive)
          a - dataset with accuracy notes
          A - archival location (no direct access to remote dataset)
          D - deleted from campaign archive (with --show_deleted option)
""",
        )
        parser_info.add_argument("-r", "--list-replicas", help="List replicas for each dataset", action="store_true")
        parser_info.add_argument(
            "-f", "--list-files", help="List files embedded in campaign archive", action="store_true"
        )
        parser_info.add_argument("-d", "--show-deleted", help="Show deleted entries", action="store_true")
        parser_info.add_argument("-c", "--show-checksum", help="Show checksums of files", action="store_true")
        parsers["info"] = parser_info

        # parser for the "dataset" command
        parser_dataset = argparse.ArgumentParser(
            prog=f"{prog} <archive> dataset",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Add one or more datasets to the archive. Datasets can be valid HDF5 or ADIOS2-BP files.
A temporary file is created from HDF5 files so one must have write access to /tmp.
""",
        )
        parsers["dataset"] = parser_dataset
        parser_dataset.add_argument("files", nargs="+", help="add ADIOS/HDF5 files manually")
        parser_dataset.add_argument("--name", "-n", default=None, help="Representation name in the campaign hierarchy")

        # parser for the "text" command
        parser_text = argparse.ArgumentParser(
            prog=f"{prog} <archive> text",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Add one or more text files to the archive. They are always stored in the archive,
so be mindful about the size of the resulting archive. Text is stored compressed.
""",
        )
        parsers["text"] = parser_text
        parser_text.add_argument("files", nargs="+", help="add text files manually")
        parser_text.add_argument("--name", "-n", default=None, help="Representation name in the campaign hierarchy")
        parser_text.add_argument("--store", "-s", help="Store image in campaign", action="store_true")

        # parser for the "image" command
        parser_image = argparse.ArgumentParser(
            prog=f"{prog} <archive> image",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Add an image file to the archive (remote reference by default).
Multiple files with different resolutions can represent an image --name in the archive.
The archive can '--store' directly the image file, or store a --thumbnail with X,Y size.
""",
        )
        parsers["image"] = parser_image
        parser_image.add_argument("file", help="image file")
        parser_image.add_argument("--name", "-n", default=None, help="Representation name in the campaign hierarchy")
        parser_image.add_argument("--store", "-s", help="Store image in campaign", action="store_true")
        parser_image.add_argument(
            "--thumbnail",
            nargs=2,
            default=None,
            type=int,
            metavar=("x", "y"),
            help="Store a resized image with resolution of x-by-y and refer to original",
        )

        # parser for the "add-archival-storage" command
        parser_addarchive = argparse.ArgumentParser(
            prog=f"{prog} <archive>  add-archival-storage",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Record an archival storage of <system> type to the list of <host>/<directory>.
Attach a --note that contains specific information on how to access the archival storage.
The <host> is a short name used to identify hosts in the campaign. Use --longhostname
to record the full hostname.
If the archive is a TAR file, add the filename of the TAR file separately in <tarfilename>.
If the TAR file is on a remotely accessible system that allows reading chunks from it,
you can generate an index of the contained files, and the datasets in the campaign archive
will be pointing to specific offsets in the TAR file.
""",
        )
        parsers["add-archival-storage"] = parser_addarchive
        parser_addarchive.add_argument(
            "system",
            choices=["Kronos", "HPSS", "fs", "https", "S3"],
            help="Name of archival system of this location",
        )
        parser_addarchive.add_argument("host", help="Archival host's name", type=str)
        parser_addarchive.add_argument("directory", help="Archival host's directory", type=str)
        parser_addarchive.add_argument("tarfilename", nargs="?", help="TAR file in directory")
        parser_addarchive.add_argument("tarfileidx", nargs="?", help="Index for TAR file")
        parser_addarchive.add_argument("--longhostname", metavar="str", help="Optional long host name")
        parser_addarchive.add_argument("--note", metavar="fname", help="Optional notes file")

        # parser for the "archived" command
        parser_archive = argparse.ArgumentParser(
            prog=f"{prog} <archive>  archived",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Indicate that a dataset/replica is copied/moved to an archival storage.
A new replica is created, pointing to the archival host/directory.
If the replica's relative path under the archival directory is different
from the original replica's path, specify it with <newpath>.
--move will delete the original replica.
Use info -r to list replicas in the campaign archive (and also to find the
archival directory's id).
If there are more than one, conflicting, replicas then --repid must be used
to indicate which version was archived.
If a directory has multiple archives (e.g. ./ and a TAR file), use --archiveid ID,
which is the second integer in the listing of directories under an archive directory
""",
        )
        parsers["archived"] = parser_archive
        parser_archive.add_argument("name", help="Name of dataset", type=str)
        parser_archive.add_argument("dirid", help="Archival host's directory ID", type=int)
        parser_archive.add_argument(
            "--archiveid",
            help="Optional archive ID if there are more than one archives in the same directory",
            type=int,
        )
        parser_archive.add_argument(
            "--newpath", help="Replica's new relative path under archival directory", type=str, metavar="str"
        )
        parser_archive.add_argument("--replica", help="Replica ID", metavar="id", type=int)
        parser_archive.add_argument("--move", help="Delete original replica", action="store_true")

        # parser for the "time-series" command
        parser_timeseries = argparse.ArgumentParser(
            prog=f"{prog} <archive>  time-series",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Organize datasets into a time-series. Enumerate the datasets in the order
they should be presented. Subsequent calls with the same <name> will add
datasets to the list of existing datasets, unless --replace is given.
""",
        )
        parsers["time-series"] = parser_timeseries
        parser_timeseries.add_argument("name", help="Name of time series", type=str)
        parser_timeseries.add_argument("dataset", nargs="*", help="Datasets in order", type=str)
        parser_timeseries.add_argument("--replace", help="Overwrite existing time-series", action="store_true")
        parser_timeseries.add_argument(
            "--remove", help="Remove the series definition (not the datasets)", action="store_true"
        )

        # parser for the "upgrade" command
        parser_upgrade = argparse.ArgumentParser(
            prog=f"{prog} <archive>  upgrade",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""
Upgrade an archive's format from its current version to the next available version.
New columns in the database will be filled with empty/null values.
One may need to run upgrade several times to reach the newest format.
""",
        )
        parsers["upgrade"] = parser_upgrade
        return parsers

    def remove_prev_args(self, command: str, args: argparse.Namespace):
        if command == "dataset":
            del args.name
            del args.files
        elif command == "text":
            del args.name
            del args.files
            del args.store
        elif command == "image":
            del args.name
            del args.file
            del args.store
            del args.thumbnail
        elif command == "info":
            del args.show_deleted
            del args.show_checksum
            del args.list_replicas
            del args.list_files
        elif command == "delete":
            del args.name
            del args.uuid
            del args.replica
            del args.campaign
        elif command == "add-archival-storage":
            del args.host
            del args.system
            del args.directory
            del args.tarfilename
            del args.tarfileidx
        elif command == "archived":
            del args.name
            del args.dirid
            del args.archiveid
            del args.newpath
            del args.replica
            del args.move
        elif command == "time-series":
            del args.name
            del args.dataset
            del args.replace
            del args.remove




import sys
import argparse
import importlib


def ArgParse():
    if ('--help' in sys.argv) and (sys.argv[1] == '--help'):
        add_help = True
    else:
        add_help = False
    parser = argparse.ArgumentParser(add_help=add_help, prog="hpc_campaign")

    parser.add_argument(
        "subcmd",
        help='Sub command',
        choices=[
            "cache",
            "connector",
            "genkey",
            "hdf5_metadata",
            "list",
            "manager",
            "taridx"
        ]
    )

    known, unknown = parser.parse_known_args()

    return known.subcmd, unknown


def main():

    subcmd, args = ArgParse()
    prog = "hpc_campaign {0}".format(subcmd)

    '''
    exec(
        'from .{0} import main as cmd'.format(subcmd),
        globals()
    )

    cmd(args=args, prog=prog)
    '''

    # Dynamically import the module using importlib.import_module()
    aliased_module = importlib.import_module(".{0}".format(subcmd), package="hpc_campaign")

    # Run main()
    aliased_module.main(args=args, prog=prog)


if __name__ == "__main__":

    main()
